
from blackfynn.api.base import APIBase

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Compute
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class ComputeAPI(APIBase):
    """
    Interface for task/workflow objects in Blackfynn
    """
    base_uri = '/compute'
    name = 'compute'

    def __init__(self, session):
        super(self.__class__, self).__init__(session)
        self.specs = ComputeSpecsAPI(session)
        self.instances = ComputeInstancesAPI(session)

    def trigger(self, org, name, inputs):
        """
        Trigger a workflow instance.

        inputs = {'var_name': {'package_id': 'something', 'type': 'package'}}
        """
        path = self._uri('/trigger/{org}/{name}',org=org, name=name)
        return self._post(path, data=inputs)


class ComputeInstancesAPI(APIBase):
    """
    Interface for task/workflow objects on Blackfynn platform
    """
    base_uri = '/compute/instances'
    name = 'compute_instances'
    _states = ['ready','initiated','running','completed']

    def set_workflow_state(self, wf_id, state, status=None):
        self._check_state(state)

        if status is None:
            status = state.title()
        path = self._uri('/workflows/{id}',id=wf_id)
        state_str = 'WorkflowInstance{}'.format(state.title())
        data = dict(
            state=dict(type=state_str),
            status=status
        )
        return self._put(endpoint, data=data)

    def set_task_state(self, wf_id, task_id):
        self._check_state(state)

        path = self._uri( '/workflows/{wf}/tasks/{task}', wf=wf_id, task=task_id)
        state_str = 'TaskInstance{}'.format(state.title())
        data = dict(
            state=dict(type=state_str),
            status=status
        )
        return self._put(path, data=data)

    def workflow_ready(self, wf_id):
        return self.set_workflow_state(wf_id, 'Ready')

    def workflow_initiated(self, wf_id):
        return self.set_workflow_state(wf_id, 'Initiated')

    def workflow_running(self, wf_id):
        return self.set_workflow_state(wf_id, 'Running')

    def workflow_complete(self, wf_id):
        return self.set_workflow_state(wf_id, 'Completed')

    def workflow_failed(self, wf_id):
        return self.set_task_state(wf_id, 'Failed')

    def task_ready(self, wf_id, task_id):
        return self.set_task_state(wf_id, task_id, 'Ready')

    def task_initiated(self, wf_id, task_id):
        return self.set_task_state(wf_id, task_id, 'Initiated')

    def task_running(self, wf_id, task_id):
        return self.set_task_state(wf_id, task_id, 'Running')

    def task_complete(self, wf_id, task_id):
        return self.set_task_state(wf_id, task_id, 'Completed')

    def task_failed(self, wf_id, task_id):
        return self.set_task_state(wf_id, task_id, 'Failed')

    def _check_state(self, state):
        if state.lower() not in self._states:
            raise Exception('State not acceptable: {}'.format(state))

class ComputeSpecsAPI(APIBase):
    """
    Interface for task/workflow objects on Blackfynn platform
    """
    base_uri = '/compute/specs'
    name = 'compute_specs'

    def _task_spec(self, org, name):
        return self._uri('/tasks/{org}/{name}',org=org, name=name)

    def _workflow_spec(self, org, name):
        return self._uri('/workflows/{org}/{name}',org=org, name=name)

    def get_task(self, org, name):
        return self._get(self._task_spec(org,name))

    def create_task(self, org, name, spec):
        resp = self._post(self._task_spec(org,name), data=spec)

    def delete_task(self, org, name):
        return self._del(self._task_spec(org,name))

    def get_workflow(self, org, name):
        return  self._get(self._workflow_spec(org,name))

    def create_workflow(self, org, name, spec):
        return self._post(self._workflow_spec(org,name), data=spec)

    def delete_workflow(self, org, name):
        return self._del(self._workflow_spec(org,name))

    def tasks(self):
        return self._get('/tasks')

    def workflows(self, **kwargs):
        return self._get('/workflows', data=kwargs)





'''
usage:
  bf cache [options] clear
  bf cache [options] compact

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt

def main():
    args = docopt(__doc__)

    from blackfynn.cache import get_cache
    cache = get_cache(init=False)
	
    if args['clear']:
        print "Clearing cache..."
        cache.clear()
        print "Cache cleared."
    elif args['compact']:
        print 'Compacting cache...'
        cache.start_compaction(async=False)
        print 'Cache compaction done.'



'''usage:
  bf [options] [<command>] [<args>...]

Available commands:
  use               Set your current working dataset
  init              Initialize a new dataset
  search            Search across your datasets
  datasets          List your datasets
  organizations     List the organizations you belong to

  collaborators     List the collaborators of the current working dataset
  share             Share a dataset with users, teams, or your organization
  unshare           Revoke access to the a dataset from users, teams, or your organization

  props             Add/remove/modify a package or collection's properties
  move              Move a package or collection
  where             Show path to package or collection
  append            Append data to a package
  rename            Rename a package or collection
  delete            Delete a package or collection
  create            Create a collection
  get               Get the contents of a package, collection, or dataset

  upload            Upload file(s) or directory

  status            Display connection status
  cache             Perform cache operations
  profile           Profile management


global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt
import os

import blackfynn
from blackfynn import Blackfynn, settings

def blackfynn_cli():
    args = docopt(__doc__,
                  version='bf version {}'.format(blackfynn.__version__),
                  options_first=True)

    #Display warning message if config.ini is not found
    if args['<command>'] != 'profile':
        if not os.path.exists(settings.config_file):
            print("\033[31m* Warning: No config file found, run 'bf profile' to start the setup assistant\033[0m")

    #Try to use profile specified by --profile, exit if invalid
    try:
        if args['--profile'] is not None:
            settings.use_profile(args['--profile'])
    except Exception, e:
        exit(e)

    #Try to use dataset specified by --dataset, exit if invalid
    try:
        if args['--dataset'] is not None:
            from cli_utils import get_client
            bf = get_client()
            dataset = bf.get_dataset(args['--dataset'])
            settings.use_dataset(dataset.id)
    except Exception, e:
        exit(e)

    if args['<command>'] == 'status':
        import bf_status
        bf_status.main()
    elif args['<command>'] == 'use':
        import bf_use
        bf_use.main()
    elif args['<command>'] == 'init':
        import bf_init
        bf_init.main()
    elif args['<command>'] in ['datasets', 'ds']:
        import bf_datasets
        bf_datasets.main()
    elif args['<command>'] in ['organizations', 'orgs']:
        import bf_organizations
        bf_organizations.main()
    elif args['<command>'] in ['share', 'unshare', 'collaborators']:
        import bf_share
        bf_share.main()
    elif args['<command>'] == 'cache':
        import bf_cache
        bf_cache.main()
    elif args['<command>'] == 'create':
        import bf_create
        bf_create.main()
    elif args['<command>'] == 'delete':
        import bf_delete
        bf_delete.main()
    elif args['<command>'] == 'move':
        import bf_move
        bf_move.main()
    elif args['<command>'] == 'rename':
        import bf_rename
        bf_rename.main()
    elif args['<command>'] == 'props':
        import bf_props
        bf_props.main()
    elif args['<command>'] == 'get':
        import bf_get
        bf_get.main()
    elif args['<command>'] == 'where':
        import bf_where
        bf_where.main()
    elif args['<command>'] == 'upload':
        import bf_upload
        bf_upload.main()
    elif args['<command>'] == 'append':
        import bf_append
        bf_append.main()
    elif args['<command>'] == 'search':
        import bf_search
        bf_search.main()
    elif args['<command>'] == 'profile':
        import bf_profile
        bf_profile.main()
    elif args['<command>'] in ['help',None]:
        print(__doc__.strip('\n'))
        return
    else:
        exit("Invalid command: '{}'\nSee 'bf help' for available commands".format(args['<command>']))



# -*- coding: utf-8 -*-

import os
import re
from uuid import uuid4
from blackfynn import settings
from blackfynn.utils import (
    infer_epoch, get_data_type, value_as_type, usecs_to_datetime
)
import datetime
import requests
import numpy as np
import dateutil.parser

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Helpers
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def get_package_class(data):
    """
    Determines package type and returns appropriate class.
    """
    content = data.get('content', data)
    if 'packageType' not in content:
        p = Dataset
    else:
        ptype = content['packageType'].lower()
        if ptype == 'collection':
            p = Collection
        elif ptype == 'timeseries':
            p = TimeSeries
        elif ptype == 'tabular':
            p = Tabular
        elif ptype == 'dataset':
            p = Dataset
        else:
            p = DataPackage

    return p

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Basics
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class Property(object):
    """
    Property of a blackfynn object. 

    Args:
        key (str): the key of the property
        value (str,number): the value of the property

        fixed (bool): if true, the value cannot be changed after the property is created
        hidden (bool): if true, the value is hidden on the platform
        category (str): the category of the property, default: "Blackfynn"
        data_type (str): one of 'string', 'integer', 'double', 'date', 'user'

    """
    _data_types = ['string', 'integer', 'double', 'date', 'user', 'boolean']
    def __init__(self, key, value, fixed=False, hidden=False, category="Blackfynn", data_type=None):
        self.key = key
        self.fixed = fixed
        self.hidden = hidden
        self.category = category

        if data_type is None or (data_type.lower() not in self._data_types):
            dt,v = get_data_type(value)
            self.data_type = dt
            self.value = v
        else:
            self.data_type = data_type
            self.value = value_as_type(value, data_type.lower())

    def as_dict(self):
        """
        Representation of instance as dictionary, used when calling API.
        """
        return {
            "key": self.key,
            "value": str(self.value), # value needs to be string :-(
            "dataType": self.data_type,
            "fixed": self.fixed,
            "hidden": self.hidden,
            "category": self.category
        }

    @classmethod
    def from_dict(cls, data, category='Blackfynn'):
        """
        Create an instance from dictionary, used when handling API response.
        """
        return cls(
            key=data['key'],
            value=data['value'],
            category=category,
            fixed=data['fixed'],
            hidden=data['hidden'],
            data_type=data['dataType']
        )

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return u"<Property key='{}' value='{}' type='{}' category='{}'>" \
                    .format(self.key, self.value, self.data_type, self.category)


def _get_all_class_args(cls):
    # possible class arguments
    if cls == object:
        return set()
    class_args = set()
    for base in cls.__bases__:
        # get all base class argument variables
        class_args.update(_get_all_class_args(base))
    # return this class and all base-class variables
    class_args.update(cls.__init__.__func__.func_code.co_varnames)
    return class_args


class BaseNode(object):
    """
    Base class to serve all objects
    """
    _api = None
    _object_key = 'content'

    def __init__(self, id=None):
        self.id = id


    @classmethod
    def from_dict(cls, data, api=None, object_key=None):
        # which object_key are we going to use?
        if object_key is not None:
            obj_key = object_key
        else:
            obj_key = cls._object_key

        # validate obj_key
        if obj_key == '' or obj_key is None:
            content = data
        else:
            content = data[obj_key]

        class_args = _get_all_class_args(cls)

        # find overlapping keys
        kwargs = {}
        thing_id = content.pop('id')
        for k,v in content.iteritems():
            # check lower case var names
            if k.lower() in class_args:
                kwargs.update({k.lower(): v})
            # check camelCase --> camel_case
            kk = re.sub(r'[A-Z]', lambda x: '_'+x.group(0).lower(), k)
            if kk in class_args:
                kwargs.update({kk: v})
            # check s3case --> s3_case
            kk = re.sub(r'[0-9]', lambda x: x.group(0)+'_', k)
            if kk in class_args:
                kwargs.update({kk: v})

        # init class with args
        item = cls.__new__(cls)
        cls.__init__(item, **kwargs)

        item.id = thing_id

        if api is not None:
            item._api = api
            item._api.core.set_local(item)

        return item

    def __eq__(self, item):
        if self.exists and item.exists:
            return self.id == item.id
        else:
            return self is item

    @property
    def exists(self):
        """ 
        Whether or not the instance of this object exists on the platform.
        """
        return self.id is not None

    def _check_exists(self):
        if not self.exists:
            raise Exception('Object must be created on the platform before method is called.') 

    def __str__(self):
        return self.__repr__()


class BaseDataNode(BaseNode):
    """
    Base class to serve all "data" node-types on platform, e.g. Packages and Collections.
    """
    _type_name = 'packageType'
    def __init__(self, name, type,
            parent=None,
            owner_id=None,
            dataset_id=None,
            id=None,
            provenance_id=None):

        super(BaseDataNode, self).__init__(id=id)
        self.name = name
        self._properties = {}
        if isinstance(parent, basestring) or parent is None:
            self.parent = parent
        elif isinstance(parent, Collection):
            self.parent = parent.id
        else:
            raise Exception("Invalid parent {}".format(parent))
        self.type = type
        self.dataset = dataset_id
        self.owner_id = owner_id
        self.provenance_id = provenance_id

        #: str: the state of the object, ``READY``, ``UNAVAILABLE``, or ``ERROR``
        self.state = None

    def update_properties(self):
        self._api.data.update_properties(self)

    def _set_properties(self, *entries):
        # Note: Property is stored as dict of key:properties-entry to enable
        #       over-write of properties values based on key
        for entry in entries:
            assert type(entry) is Property, "Properties wrong type"
            if entry.category not in self._properties:
                self._properties[entry.category] = {}
            self._properties[entry.category].update({entry.key:entry})

    def add_properties(self, *entries):
        """ 
        Add properties to object.

        Args:
            entries (list): list of Property objects to add to this object

        """
        self._set_properties(*entries)

        # update on platform
        if self.exists:
            self.update_properties()

    def insert_property(self, key, value, fixed=False, hidden=False, category="Blackfynn", data_type=None):
        """
        Add property to object using simplified interface.

        Args:
            key (str): the key of the property
            value (str,number): the value of the property

            fixed (bool): if true, the value cannot be changed after the property is created
            hidden (bool): if true, the value is hidden on the platform
            category (str): the category of the property, default: "Blackfynn"
            data_type (str): one of 'string', 'integer', 'double', 'date', 'user'

        Note:
            This method is being depreciated in favor of ``set_property()`` method (see below).

        """
        return self.set_property(
            key=key,
            value=value,
            fixed=fixed,
            hidden=hidden,
            category=category,
            data_type=data_type
        )

    def set_property(self, key, value, fixed=False, hidden=False, category="Blackfynn", data_type=None):
        """
        Add property to object using simplified interface.

        Args:
            key (str): the key of the property
            value (str,number): the value of the property

            fixed (bool): if true, the value cannot be changed after the property is created
            hidden (bool): if true, the value is hidden on the platform
            category (str): the category of the property, default: "Blackfynn"
            data_type (str): one of 'string', 'integer', 'double', 'date', 'user'

        """
        self._set_properties(
            Property(
                key=key,
                value=value,
                fixed=fixed,
                hidden=hidden,
                category=category,
                data_type=data_type)
        )
        # update on platform, if possible
        if self.exists:
            self.update_properties()

    @property
    def properties(self):
        """
        Returns a list of properties attached to object.
        """
        props = []
        for category in self._properties.values():
            props.extend(category.values())
        return props

    def get_property(self, key, category='Blackfynn'):
        """
        Returns a single property for the provided key, if available

        Args:
            key (str): key of the desired property
            category (str, optional): category of property

        Returns:
            object of type ``Property``

        Example::

            pkg.set_property('quality', 85.0)
            pkg.get_property('quality')

        """
        return self._properties[category].get(key, None)

    def remove_property(self, key, category='Blackfynn'):
        """
        Removes property of key ``key`` and category ``category`` from the object.

        Args:
            key (str): key of property to remove
            category (str, optional): category of property to remove

        """
        if key in self._properties[category]:
            # remove by setting blank
            self._properties[category][key].value = ""
            # update remotely
            self.update_properties()
            # get rid of it locally
            self._properties[category].pop(key)

    def update(self, **kwargs):
        """
        Updates object on the platform (with any local changes) and syncs
        local instance with API response object.

        Exmple::

            pkg = bf.get('N:package:1234-1234-1234-1234')
            pkg.name = "New name"
            pkg.update()

        """
        self._check_exists()
        r = self._api.core.update(self, **kwargs)
        self.__dict__.update(r.__dict__)

    def delete(self):
        """
        Delete object from platform.
        """
        self._check_exists()
        r = self._api.core.delete(self)
        self.id = None

    def set_ready(self, **kwargs):
        """
        Set's the package's state to ``READY``
        """
        self.state = "READY"
        return self.update(**kwargs)

    def set_unavailable(self):
        """
        Set's the package's state to ``UNAVAILABLE``
        """
        self._check_exists()
        self.state = "UNAVAILABLE"
        return self.update()

    def set_error(self):
        """
        Set's the package's state to ``ERROR``
        """
        self._check_exists()
        self.state = "ERROR"
        return self.update()

    def as_dict(self):
        d = {
            "name": self.name,
            self._type_name: self.type,
            "properties": [
                m.as_dict() for m in self.properties
            ]
        }

        for k in ['parent', 'state', 'dataset']:
            kval = self.__dict__.get(k, None)
            if hasattr(self, k) and kval is not None:
                d[k] = kval
                
        if self.provenance_id is not None:
            d["provenanceId"] = self.provenance_id

        return d

    @classmethod
    def from_dict(cls, data, *args, **kwargs):
        item = super(BaseDataNode,cls).from_dict(data, *args, **kwargs)

        try:
            item.state = data['content']['state']
        except:
            pass

        # parse, store parent (ID only)
        if 'parent' in data:
            parent = data['parent']
            if isinstance(parent, basestring):
                item.parent = parent
            else:
                pkg_cls = get_package_class(parent) 
                p = pkg_cls.from_dict(parent, *args, **kwargs)
                item.parent = p.id

        def cls_add_property(prop):
            cat = prop.category
            if cat not in item._properties:
                item._properties[cat] = {}
            item._properties[cat].update({prop.key: prop})

        # parse properties
        if 'properties' in data:
            for entry in data['properties']:
                if 'properties' not in entry:
                    # flat list of properties: [entry]
                    prop = Property.from_dict(entry, category=entry['category'])
                    cls_add_property(prop)
                else:
                    # nested properties list [ {category,entry} ]
                    category = entry['category']
                    for prop_entry in entry['properties']:
                        prop = Property.from_dict(prop_entry, category=category)
                        cls_add_property(prop)

        return item


class BaseCollection(BaseDataNode):
    """
    Base class used for both ``Dataset`` and ``Collection``.
    """
    def __init__(self, name, package_type, **kwargs):
        super(BaseCollection, self).__init__(name, package_type, **kwargs)

        # items is None until an API response provides the item objects 
        # to be parsed, which then updates this instance.
        self._items = None

    def add(self, *items):
        """
        Add items to the Collection/Dataset.
        """
        self._check_exists()
        for item in items:
            # initialize if need be
            if self._items is None:
                self._items = [] 
            if isinstance(self, Dataset):
                item.parent = None
                item.dataset = self.id
            elif hasattr(self, 'dataset'):
                item.parent = self.id
                item.dataset = self.dataset

            # create, if not already created
            new_item = self._api.core.create(item)
            item.__dict__.update(new_item.__dict__)

            # add item
            self._items.append(item)

    def remove(self, *items):
        """
        Removes items, where items can be an object or the object's ID (string).
        """
        self._check_exists()
        for item in items:
            if item not in self._items:
                raise Exception('Cannot remove item, not in collection:{}'.format(item))

        self._api.data.delete(*items)
        # force refresh
        self._items = None

    @property
    def items(self):
        """
        Get all items inside Dataset/Collection (i.e. non-nested items). 

        Note:
            You can also iterate over items inside a Dataset/Colleciton without using ``.items``::

                for item in my_dataset:
                    print "item name = ", item.name

        """
        self._check_exists()
        if self._items is None:
            new_self = self._get_method(self)
            new_items = new_self._items
            self._items = new_items if new_items is not None else []

        return self._items

    @property
    def _get_method(self):
        pass

    def print_tree(self, indent=0):
        """
        Prints a tree of **all** items inside object.
        """
        self._check_exists()
        print u'{}{}'.format(' '*indent, self)
        for item in self.items:
            if isinstance(item, BaseCollection):
                item.print_tree(indent=indent+2)
            else:
                print u'{}{}'.format(' '*(indent+2), item)

    def get_items_by_name(self, name):
        """
        Get an item inside of object by name (if match is found).

        Args:
            name (str): the name of the item

        Returns:
            list of matches

        Note:
            This only works for **first-level** items, meaning it must exist directly inside the current object;
            nested items will not be returned.

        """
        self._check_exists()
        # note: non-hierarchical
        return filter(lambda x: x.name==name, self.items)

    def get_items_names(self):
        self._check_exists()
        return map(lambda x: x.name, self.items)

    def upload(self, *files):
        """
        Upload files into current object.

        Args:
            files: list of local files to upload.

        Example::

            my_collection.upload('/path/to/file1.nii.gz', '/path/to/file2.pdf')

        """
        self._check_exists()
        return self._api.io.upload_files(self, files, append=False)

    def create_collection(self, name):
        """
        Create a new collection within the current object. Collections can be created within 
        datasets and within other collections.

        Args:
            name (str): The name of the to-be-created collection

        Returns:
            The created ``Collection`` object.

        Example::

              from blackfynn import Blackfynn()
              
              bf = Blackfynn()
              ds = bf.get_dataset('my_dataset')
              
              # create collection in dataset
              col1 = ds.create_collection('my_collection')

              # create collection in collection
              col2 = col1.create_collection('another_collection')
        
        """
        c = Collection(name)
        self.add(c)
        return c

    # sequence-like method
    def __getitem__(self, i):
        self._check_exists()
        return self.items[i]

    # sequence-like method
    def __len__(self):
        self._check_exists()
        return len(self.items)

    # sequence-like method
    def __delitem__(self, key):
        self._check_exists()
        self.remove(key)

    def __iter__(self):
        self._check_exists()
        for item in self.items:
            yield item

    # sequence-like method
    def __contains__(self, item):
        """
        Tests if item is in the collection, where item can be either
        an object's ID (string) or an object's instance.
        """
        self._check_exists()
        if isinstance(item, basestring):
            some_id = self._api.data._get_id(item)
            item_ids = [x.id for x in self.items]
            contains = some_id in item_ids
        elif self._items is None:
            return False
        else:
            return item in self._items

        return contains

    @classmethod
    def from_dict(cls, data, *args, **kwargs):
        item = super(BaseCollection, cls).from_dict(data, *args, **kwargs)
        children = []
        if 'children' in data:
            for child in data['children']:
                pkg_cls = get_package_class(child)
                kwargs['api'] = item._api
                pkg = pkg_cls.from_dict(child, *args, **kwargs)
                children.append(pkg)
        item.add(*children)

        return item

    def __repr__(self):
        return u"<BaseCollection name='{}' id='{}'>".format(self.name, self.id)


class DataPackage(BaseDataNode):
    """
    DataPackage is the core data object representation on the platform.

    Args:
        name (str):          The name of the data package
        package_type (str):  The package type, e.g. 'TimeSeries', 'MRI', etc.

    Note:
        ``package_type`` must be a supported package type. See our data type
        registry for supported values.

    """

    def __init__(self, name, package_type, **kwargs):
        super(DataPackage, self).__init__(name=name, type=package_type, **kwargs)
        # local-only attribute
        self.session = None

    def set_view(self, *files):
        """
        Set the object(s) used to view the package, if not the file(s) or source(s).
        """
        self._check_exists()
        ids = self._api.packages.set_view(self, *files)
        # update IDs of file objects
        for i,f in enumerate(files):
            f.id = ids[i]

    def set_files(self, *files):
        """
        Sets the files of a DataPackage. Files are typically modified 
        source files (e.g. converted to a different format).
        """
        self._check_exists()
        ids = self._api.packages.set_files(self, *files)
        # update IDs of file objects
        for i,f in enumerate(files):
            f.id = ids[i]

    def set_sources(self, *files):
        """
        Sets the sources of a DataPackage. Sources are the raw, unmodified
        files (if they exist) that contains the package's data.
        """
        self._check_exists()
        ids = self._api.packages.set_sources(self, *files)
        # update IDs of file objects
        for i,f in enumerate(files):
            f.id = ids[i]

    def append_to_files(self, *files):
        """
        Append to file list of a DataPackage
        """
        self._check_exists()
        files = self._api.packages.set_files(self, *files, append=True)

    def append_to_sources(self, *files):
        """
        Appends to source list of a DataPackage.
        """
        self._check_exists()
        files = self._api.packages.set_sources(self, *files, append=True)

    @property
    def sources(self):
        """
        Returns the sources of a DataPackage. Sources are the raw, unmodified
        files (if they exist) that contains the package's data.
        """
        self._check_exists()
        return self._api.packages.get_sources(self)

    @property
    def files(self):
        """
        Returns the files of a DataPackage. Files are the possibly modified 
        source files (e.g. converted to a different format), but they could also
        be the source files themselves.
        """
        self._check_exists()
        return self._api.packages.get_files(self)

    @property
    def view(self):
        """
        Returns the object(s) used to view the package. This is typically a set of
        file objects, that may be the DataPackage's sources or files, but could also be
        a unique object specific for the viewer.
        """
        self._check_exists()
        return self._api.packages.get_view(self)

    @classmethod
    def from_dict(cls, data, *args, **kwargs):
        item = super(DataPackage, cls).from_dict(data, *args, **kwargs)

        # parse objects
        if 'objects' in data:
            for otype in ['sources','files','view']:
                if otype not in data['objects']:
                    continue
                odata = data['objects'][otype]
                item.__dict__[otype] = [File.from_dict(x) for x in odata]

        return item

    @classmethod
    def from_id(cls, id):
        return self._api.packages.get(id)

    def __repr__(self):
        return u"<DataPackage name='{}' id='{}'>".format(self.name, self.id)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Files
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class File(BaseDataNode):
    """
    File node on the Blackfynn platform. Points to some S3 location.

    Args:
        name (str):      Name of the file (without extension)
        s3_key (str):    S3 key of file
        s3_bucket (str): S3 bucket of file
        file_type (str): Type of file, e.g. 'MPEG', 'PDF'

    Note:
        ``file_type`` must be a supported file type. See our file type registry
        for a list of supported file types.


    """
    _type_name = 'fileType'

    def __init__(self, name, s3_key, s3_bucket, file_type, pkg_id=None, **kwargs):
        super(File, self).__init__(name, type=file_type, **kwargs)

        # data
        self.s3_key = s3_key
        self.s3_bucket = s3_bucket
        self.pkg_id = pkg_id
        self.local_path = None

    def as_dict(self):
        d = super(File, self).as_dict()
        d.update({
            "s3bucket": self.s3_bucket,
            "s3key": self.s3_key
        })
        d.pop('parent', None)
        props = d.pop('properties')
        return {
            'objectType': 'file',
            'content': d,
            'properties': props
        }

    @property
    def url(self):
        """
        The presigned-URL of the file.
        """
        self._check_exists()
        return self._api.packages.get_presigned_url_for_file(self.pkg_id, self.id)

    def download(self, destination):
        """
        Download the file.

        Args:
            destination (str): path for downloading; can be absolute file path, 
                               prefix or destination directory.

        """
        if self.type=="DirectoryViewerData":
            raise NotImplementedError("Downloading S3 directories is currently not supported")

        if os.path.isdir(destination):
            # destination dir
            f_local = os.path.join(destination, os.path.basename(self.s3_key))
        if '.' not in os.path.basename(destination):
            # destination dir + prefix
            f_local = destination + '_' + os.path.basename(self.s3_key)
        else:
            # exact location
            f_local = destination

        r = requests.get(self.url, stream=True)
        with open(f_local, 'wb') as f:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk: f.write(chunk)

        # set local path
        self.local_path = f_local

        return f_local

    def __repr__(self):
        return u"<File name='{}' type='{}' key='{}' bucket='{}' id='{}'>" \
                    .format(self.name, self.type, self.s3_key, self.s3_bucket, self.id)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Time series
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class TimeSeries(DataPackage):
    """
    Represents a timeseries package on the platform. TimeSeries packages
    contain channels, which contain time-dependent data sampled at some
    frequency.

    Args:
        name:  The name of the timeseries package
        
    """
    def __init__(self, name, **kwargs):
        kwargs.pop('package_type', None)
        super(TimeSeries,self).__init__(name=name, package_type="TimeSeries", **kwargs)


    def streaming_credentials(self):
        self._check_exists()
        return self._api.timeseries.get_streaming_credentials(self)

    @property
    def start(self):
        """
        The start time of time series data (over all channels)
        """
        self._check_exists()
        return sorted([x.start for x in self.channels])[0]

    @property
    def end(self):
        """
        The end time (in usecs) of time series data (over all channels)
        """
        self._check_exists()
        return sorted([x.end for x in self.channels])[-1]

    def limits(self):
        """
        Returns time limit tuple (start, end) of package.
        """
        channels = self.channels
        start = sorted([x.start for x in channels])[0]
        end   = sorted([x.end   for x in channels])[-1]
        return start, end

    @property
    def channels(self):
        """
        Returns list of Channel objects associated with package.

        Note:
            This is a dynamically generated property, so every call will make an API request.

            Suggested usage::

                channels = ts.channels
                for ch in channels:
                    print ch
            
            This will be much slower, as the API request is being made each time.::

                for ch in ts.channels:
                    print ch

        """
        self._check_exists()
        # always dynamically return channel list
        return self._api.timeseries.get_channels(self)

    def get_channel(self, channel):
        """
        Get channel by ID.

        Args:
            channel (str): ID of channel
        """
        self._check_exists()
        return self._api.timeseries.get_channel(self, channel)

    def add_channels(self, *channels):
        """
        Add channels to TimeSeries package.

        Args:
            channels: list of Channel objects.

        """
        self._check_exists()
        for channel in channels:
            ch = self._api.timeseries.create_channel(self, channel)
            channel.__dict__.update(ch.__dict__)

    def remove_channels(self, *channels):
        """
        Remove channels from TimeSeries package.

        Args:
            channels: list of Channel objects or IDs
        """
        self._check_exists()
        for channel in channels:
            self._api.timeseries.delete_channel(channel)
            channel.id = None
            channel._pkg = None

    # ~~~~~~~~~~~~~~~~~~
    # Data 
    # ~~~~~~~~~~~~~~~~~~
    def get_data(self, start=None, end=None, length=None, channels=None, use_cache=settings.use_cache):
        """
        Get timeseries data between ``start`` and ``end`` or ``start`` and ``start + length`` 
        on specified channels (default all channels).

        Args:
            start (optional): start time of data (usecs or datetime object)
            end (optional): end time of data (usecs or datetime object)
            length (optional): length of data to retrieve, e.g. '1s', '5s', '10m', '1h'
            channels (optional): list of channel objects or IDs, default all channels.

        Note:
            Data requests will be automatically chunked and combined into a single Pandas
            DataFrame. However, you must be sure you request only a span of data that
            will properly fit in memory. 

            See ``get_data_iter`` for an iterator approach to timeseries data retrieval.

        Examples:

            Get 5 seconds of data from start over all channels::

                data = ts.get_data(length='5s')

            Get data betwen 12345 and 56789 (representing usecs since Epoch)::

                data = ts.get_data(start=12345, end=56789)

            Get first 10 seconds for the first two channels::

                data = ts.get_data(length='10s', channels=ts.channels[:2])

        """
        self._check_exists()
        return self._api.timeseries.get_ts_data(self,start=start, end=end, length=length, channels=channels, use_cache=use_cache)

    def get_data_iter(self, channels=None, start=None, end=None, length=None, chunk_size=None, use_cache=settings.use_cache):
        """
        Returns iterator over the data. Must specify **either ``end`` OR ``length``**, not both.

        Args:
            channels (optional): channels to retrieve data for (default: all)
            start: start time of data (default: earliest time available).
            end: end time of data (default: latest time avialable).
            length: some time length, e.g. '1s', '5m', '1h' or number of usecs
            chunk: some time length, e.g. '1s', '5m', '1h' or number of usecs

        Returns:
            iterator of Pandas Series, each the size of ``chunk_size``.

        """
        self._check_exists()
        return self._api.timeseries.get_ts_data_iter(self, channels=channels, start=start, end=end, length=length, chunk_size=chunk_size, use_cache = use_cache)

    def write_annotation_file(self,file,layer_names = None):
        """
        Writes all layers to a csv .bfannot file

        Args:
            file : path to .bfannot output file. Appends extension if necessary
            layer_names (optional): List of layer names to write

        """

        return self._api.timeseries.write_annotation_file(self,file,layer_names)

    def append_annotation_file(self,file):
        """
        Processes .bfannot file and adds to timeseries package.

        Args:
            file : path to .bfannot file

        """
        self._check_exists()
        return self._api.timeseries.process_annotation_file(self,file)

    def append_files(self, *files):
        self._check_exists()
        return self._api.io.upload_files(self, files, append=True)

    def stream_data(self, data):
        self._check_exists()
        return self._api.timeseries.stream_data(self, data)

    # ~~~~~~~~~~~~~~~~~~
    # Annotations
    # ~~~~~~~~~~~~~~~~~~

    @property
    def layers(self):
        """
        List of annotation layers attached to TimeSeries package.
        """
        self._check_exists()
        # always dynamically return annotation layers
        return self._api.timeseries.get_annotation_layers(self)

    def get_layer(self, id_or_name):
        """
        Get annotation layer by ID or name.

        Args:
            id_or_name: layer ID or name
        """
        self._check_exists()
        layers = self.layers
        matches = filter(lambda x: x.id==id_or_name, layers)
        if len(matches) == 0:
            matches = filter(lambda x: x.name==id_or_name, layers)

        if len(matches) == 0:
            raise Exception("No layers match criteria.")
        if len(matches) > 1:
            raise Exception("More than one layer matched criteria")
            
        return matches[0]

    def add_layer(self,layer,description=None):
        """
        Args:
            layer:   TimeSeriesAnnotationLayer object or name of annotation layer
            description (str, optional):   description of layer

        """
        self._check_exists()
        return self._api.timeseries.create_annotation_layer(self,layer=layer,description=description)

    def add_annotations(self,layer,annotations):
        """
        Args:
            layer: either TimeSeriesAnnotationLayer object or name of annotation layer.
                   Note that non existing layers will be created.
            annotations: TimeSeriesAnnotation object(s)

        Returns:
            list of TimeSeriesAnnotation objects
        """
        self._check_exists()
        cur_layer = self._api.timeseries.create_annotation_layer(self,layer=layer,description=None)
        return self._api.timeseries.create_annotations(layer=cur_layer, annotations=annotations)

    def insert_annotation(self,layer,annotation,start=None,end=None,channel_ids=None,annotation_description=None):
        """
        Insert annotations using a more direct interface, without the need for layer/annotation objects.

        Args:
            layer: str of new/existing layer or annotation layer object
            annotation: str of annotation event

            start (optional): start of annotation
            end (optional): end of annotation
            channels_ids (optional): list of channel IDs to apply annotation
            annotation_description (optional): description of annotation 

        Example:
            To add annotation on layer "my-events" across all channels::

                ts.insert_annotation('my-events', 'my annotation event')

            To add annotation to first channel::

                ts.insert_annotation('my-events', 'first channel event', channel_ids=ts.channels[0])

        """
        self._check_exists()
        cur_layer = self._api.timeseries.create_annotation_layer(self,layer=layer,description=None)
        return self._api.timeseries.create_annotation(
                layer=cur_layer,
                annotation=annotation,
                start=start,
                end=end,
                channel_ids=channel_ids,
                description=annotation_description)

    def delete_layer(self, layer):
        """
        Delete annotation layer.

        Args:
            layer: annotation layer object

        """
        self._check_exists()
        return self._api.timeseries.delete_annotation_layer(layer)

    def query_annotation_counts(self, channels, start, end, layer=None):
        """
        Get annotation counts between ``start`` and ``end``.

        Args:
            channels: list of channels to query over
            start: start time of query (datetime object or usecs from Epoch)
            end: end time of query (datetime object or usecs from Epoch)

        """
        self._check_exists()
        return self._api.timeseries.query_annotation_counts(
            channels=channels,start=start,end=end,layer=layer)

    def __repr__(self):
        return u"<TimeSeries name=\'{}\' id=\'{}\'>".format(self.name, self.id)


class TimeSeriesChannel(BaseDataNode):
    """
    TimeSeriesChannel represents a single source of time series data. (e.g. electrode)

    Args:
        name (str):                   Name of channel
        rate (float):                 Rate of the channel (Hz)
        start (optional):             Absolute start time of all data (datetime obj)
        end (optional):               Absolute end time of all data (datetime obj)
        unit (str, optional):         Unit of measurement
        channel_type (str, optional): One of 'continuous' or 'event'
        source_type (str, optional):  The source of data, e.g. "EEG"
        group (str, optional):        The channel group, default: "default"

    """
    def __init__(self, name, rate, start=0, end=0, unit='V', channel_type='continuous', source_type='unspecified', group="default", last_annot=0, spike_duration=None, **kwargs):
        self.channel_type = channel_type.upper()

        super(TimeSeriesChannel, self).__init__(name=name, type=self.channel_type,**kwargs)

        self.rate = rate
        self.unit = unit
        self.last_annot = last_annot
        self.group = group
        self.start = start
        self.end = end
        self.spike_duration = spike_duration

        self.set_property("Source Type", source_type.upper(), fixed=True, hidden=True, category="Blackfynn")

        ###  local-only
        # parent package
        self._pkg = None
        # sample period (in usecs)
        self._sample_period = 1.0e6/self.rate

    @property
    def start(self):
        """
        The start time of channel data (microseconds since Epoch)
        """
        return self._start

    @start.setter
    def start(self, start):
        self._start = infer_epoch(start)

    @property
    def start_datetime(self):
        return usecs_to_datetime(self._start)

    @property
    def end(self):
        """
        The end time (in usecs) of channel data (microseconds since Epoch)
        """
        return self._end

    @end.setter
    def end(self, end):
        self._end = infer_epoch(end)

    @property
    def end_datetime(self):
        return usecs_to_datetime(self._end)

    def _page_delta(self, page_size):
        return long((1.0e6/self.rate) * page_size)

    def update(self):
        self._check_exists()
        r = self._api.timeseries.update_channel(self)
        self.__dict__.update(r.__dict__)

    @property
    def segments(self):
        # TODO: query API to get segments
        raise NotImplementedError

    @property
    def gaps(self):
        # TODO: infer gaps from segments
        raise NotImplementedError

    def update_properties(self):
        self._api.timeseries.update_channel_properties(self)

    def get_data(self, start=None, end=None, length=None, use_cache=settings.use_cache):
        """
        Get channel data between ``start`` and ``end`` or ``start`` and ``start + length`` 

        Args:
            start     (optional): start time of data (usecs or datetime object)
            end       (optional): end time of data (usecs or datetime object)
            length    (optional): length of data to retrieve, e.g. '1s', '5s', '10m', '1h'
            use_cache (optional): whether to use locally cached data

        Returns:
            Pandas Series containing requested data for channel.

        Note:
            Data requests will be automatically chunked and combined into a single Pandas
            Series. However, you must be sure you request only a span of data that
            will properly fit in memory. 

            See ``get_data_iter`` for an iterator approach to timeseries data retrieval.

        Examples:

            Get 5 seconds of data from start over all channels::

                data = channel.get_data(length='5s')

            Get data betwen 12345 and 56789 (representing usecs since Epoch)::

                data = channel.get_data(start=12345, end=56789)
        """

        return self._api.timeseries.get_ts_data(
                ts         = self._pkg,
                start      = start,
                end        = end,
                length     = length,
                channels   = [self],
                use_cache  = use_cache)

    def get_data_iter(self, start=None, end=None, length=None, chunk_size=None, use_cache=settings.use_cache):
        """
        Returns iterator over the data. Must specify **either ``end`` OR ``length``**, not both.

        Args:
            start      (optional): start time of data (default: earliest time available).
            end        (optional): end time of data (default: latest time avialable).
            length     (optional): some time length, e.g. '1s', '5m', '1h' or number of usecs
            chunk_size (optional): some time length, e.g. '1s', '5m', '1h' or number of usecs
            use_cache  (optional): whether to use locally cached data

        Returns:
            Iterator of Pandas Series, each the size of ``chunk_size``.
        """

        return self._api.timeseries.get_ts_data_iter(
                ts         = self._pkg,
                start      = start,
                end        = end,
                length     = length,
                channels   = [self],
                chunk_size = chunk_size,
                use_cache  = use_cache)

    def as_dict(self):
        return {
            "name": self.name,
            "start": self.start,
            "end": self.end,
            "unit": self.unit,
            "rate": self.rate,
            "channelType": self.channel_type,
            "lastAnnotation": self.last_annot,
            "group": self.group,
            "spikeDuration": self.spike_duration,
            "properties": [x.as_dict() for x in self.properties]
        }

    def __repr__(self):
        return u"<TimeSeriesChannel name=\'{}\' id=\'{}\'>".format(self.name, self.id)


class TimeSeriesAnnotationLayer(BaseNode):
    """
    Annotation layer containing one or more annotations. Layers are used
    to separate annotations into logically distinct groups when applied
    to the same data package. 

    Args:
        name:           Name of the layer
        time_series_id: The TimeSeries ID which the layer applies
        description:    Description of the layer

    """
    _object_key = None

    def __init__(self, name, time_series_id, description=None, **kwargs):
        super(TimeSeriesAnnotationLayer,self).__init__(**kwargs)
        self.name = name
        self.time_series_id= time_series_id
        self.description = description

    def iter_annotations(self, window_size=10, channels=None):
        """
        Iterate over annotations according to some window size (seconds).

        Args:
            window_size (float): Number of seconds in window
            channels:            List of channel objects or IDs
        
        Yields:
            List of annotations found in current window.
        """
        self._check_exists()
        ts = self._api.core.get(self.time_series_id)
        return self._api.timeseries.iter_annotations(
            ts=ts, layer=self, channels=channels, window_size=window_size)

    def add_annotations(self, annotations):
        """
        Add annotations to layer.

        Args:
            annotations (str): List of annotation objects to add.

        """
        self._check_exists()
        return self._api.timeseries.create_annotations(layer=self, annotations=annotations)

    def insert_annotation(self,annotation,start=None,end=None,channel_ids=None,description=None):
        """
        Add annotations; proxy for ``add_annotations``. 

        Args:
            annotation (str): Annotation string
            start:            Start time (usecs or datetime)
            end:              End time (usecs or datetime)
            channel_ids:      list of channel IDs

        Returns:
            The created annotation object.
        """
        self._check_exists()
        return self._api.timeseries.create_annotation(layer=self, annotation=annotation,start=start,end=end,channel_ids=channel_ids,description=description)

    def annotations(self, start=None, end=None, channels=None):
        """
        Get annotations between ``start`` and ``end`` over ``channels`` (all channels by default).


        Args:
            start:    Start time
            end:      End time
            channels: List of channel objects or IDs

        """
        self._check_exists()
        ts = self._api.core.get(self.time_series_id)
        return self._api.timeseries.query_annotations(
            ts=ts, layer=self, channels=channels, start=start, end=end)

    def annotation_counts(self, start, end, channels=None):
        """
        The number of annotations between ``start`` and ``end`` over selected 
        channels (all by default).

        Args:
            start:    Start time
            end:      End time
            channels: List of channel objects or IDs
        """
        self._check_exists()
        ts = self._api.core.get(self.time_series_id)
        return self._api.timeseries.query_annotation_counts(
            ts=ts, layer=self, channels=channels, start=start, end=end)

    def delete(self):
        """
        Delete annotation layer.
        """
        self._check_exists()
        return self._api.timeseries.delete_annotation_layer(self)

    def as_dict(self):
        return {
            "name" : self.name,
            "description" : self.description
        }

    def __repr__(self):
        return u"<TimeSeriesAnnotationLayer name=\'{}\' id=\'{}\'>".format(self.name, self.id)


class TimeSeriesAnnotation(BaseNode):
    """
    Annotation is an event on one or more channels in a dataset

    Args:
        label (str):    The label for the annotation
        channel_ids:    List of channel IDs that annotation applies
        start:          Start time
        end:            End time
        name:           Name of annotation
        layer_id:       Layer ID for annoation (all annotations exist on a layer)
        time_series_id: TimeSeries package ID
        description:    Description of annotation

    """
    _object_key = None


    def __init__(self, label, channel_ids, start, end, name='',layer_id= None, 
                 time_series_id = None, description=None, **kwargs):
        super(TimeSeriesAnnotation,self).__init__(**kwargs)
        self.name = ''
        self.label = label
        self.channel_ids = channel_ids
        self.start = start
        self.end = end
        self.description = description
        self.layer_id = layer_id
        self.time_series_id = time_series_id

    def delete(self):
        self._check_exists()
        return self._api.timeseries.delete_annotation(annot=self)

    def as_dict(self):
        channel_ids = self.channel_ids
        if type(channel_ids) is not list:
            channel_ids = [channel_ids]
        return {
            "name" : self.name,
            "label" : self.label, 
            "channelIds": channel_ids,
            "start" : self.start, 
            "end" : self.end, 
            "description" : self.description, 
            "layer_id" : self.layer_id, 
            "time_series_id" : self.time_series_id,
        }

    def __repr__(self):
        date = datetime.datetime.fromtimestamp(self.start/1e6)
        return u"<TimeSeriesAnnotation label=\'{}\' layer=\'{}\' start=\'{}\'>".format(self.label, self.layer_id, date.isoformat())


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Tabular
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class Tabular(DataPackage):
    """
    Represents a Tabular package on the platform.

    Args:
        name: The name of the package
    """
    def __init__(self, name, **kwargs):
        kwargs.pop('package_type',None)
        super(Tabular,self).__init__(
            name=name,
            package_type="Tabular",
            **kwargs)
        self.schema = None

    def get_data(self,limit=1000, offset=0,order_by = None, order_direction='ASC'):
        """
        Get data from tabular package as DataFrame

        Args:
            limit:           Max number of rows to return (1000 default)
            offset:          Offset when retrieving rows
            order_by:        Column to order data
            order_direction: Ascending ('ASC') or descending ('DESC')

        Returns:
            Pandas DataFrame

        """
        self._check_exists()
        return self._api.tabular.get_tabular_data(self,limit=limit,offset=offset ,order_by=order_by, order_direction=order_direction)

    def get_data_iter(self, chunk_size=10000, offset=0, order_by = None, order_direction='ASC'):
        """
        Iterate over tabular data, each data chunk will be of size ``chunk_size``.
        """
        self._check_exists()
        return self._api.tabular.get_tabular_data_iter(self,chunk_size=chunk_size,offset=offset,order_by=order_by, order_direction=order_direction)

    def set_schema(self, schema):
        self.schema = schema
        # TODO: parse response
        return self._api.tabular.set_table_schema(self, schema)

    def get_schema(self):
        self._check_exists()
        # TODO: parse response
        return self._api.tabular.get_table_schema(self)

    def __repr__(self):
        return u"<Tabular name=\'{}\' id=\'{}\'>".format(self.name, self.id)


class TabularSchema(BaseNode):
    def __init__(self, name, column_schema = [], **kwargs):
        super(TabularSchema, self).__init__(**kwargs)
        self.name = name
        self.column_schema = column_schema

    @classmethod
    def from_dict(cls, data):
        column_schema = []
        for x in data['columns']:
            if 'displayName' not in x.keys():
                x['displayName'] = ''
            column_schema.append(TabularSchemaColumn.from_dict(x))

        return cls(
            name = data['name'],
            id = data['id'],
            column_schema = column_schema
           )

    def as_dict(self):
        column_schema = [dict(
            name = x.name,
            displayName = x.display_name,
            datatype = x.datatype,
            primaryKey = x.primary_key,
            internal = x.internal
        ) for x in self.column_schema]
        return column_schema

    def __repr__(self):
        return u"<TabularSchema name=\'{}\' id=\'{}\'>".format(self.name, self.id)

class TabularSchemaColumn():

    def __init__(self, name, display_name, datatype, primary_key = False, internal = False, **kwargs):
        self.name=name
        self.display_name = display_name
        self.datatype = datatype
        self.internal = internal
        self.primary_key = primary_key

    @classmethod
    def from_dict(cls, data):
        return cls(
            name = data['name'],
            display_name = data['displayName'],
            datatype = data['datatype'],
            primary_key = data['primaryKey'],
            internal = data['internal']          
        )

    def __repr__(self):
        return u"<TabularSchemaColumn name='{}' display='{}' is-primary='{}'>".format(self.name, self.display_name, self.primary_key)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# User
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class User(BaseNode):

    _object_key = ''

    def __init__(self,
            email,
            first_name,
            last_name,
            credential='',
            photo_url='',
            url='',
            authy_id=0,
            accepted_terms='',
            color=None,
            is_super_admin=False,
            *args,
            **kwargs):
        super(User, self).__init__(*args, **kwargs)

        self.email = email
        self.first_name = first_name
        self.last_name = last_name
        self.credential = credential
        self.photo_url = photo_url
        self.color = color
        self.url = url
        self.authy_id = authy_id
        self.accepted_terms = ''
        self.is_super_admin = is_super_admin

    def __repr__(self):
        return u"<User email=\'{}\' id=\'{}\'>".format(self.email, self.id)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Organizations
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class Organization(BaseNode):
    _object_key = 'organization'

    def __init__(self,
            name,
            encryption_key_id="", 
            slug=None,
            terms=None,
            *args, **kwargs):
        super(Organization, self).__init__(*args, **kwargs)

        self.name = name
        self.terms = terms
        self.encryption_key_id = encryption_key_id
        self.slug = name.lower().replace(' ','-') if slug is None else slug

    @property
    def datasets(self):
        """
        Return all datasets for user for an organization (current context).
        """
        self._check_exists()
        return self._api.datasets.get_all()

    def __repr__(self):
        return u"<Organization name=\'{}\' id=\'{}\'>".format(self.name, self.id)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Datasets
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class Dataset(BaseCollection):
    def __init__(self, name, description=None, **kwargs):
        kwargs.pop('package_type', None)
        super(Dataset, self).__init__(name, "DataSet", **kwargs)
        self.description = description or ''

        # remove things that do not apply (a bit hacky)
        for k in ("parent", "type", "set_ready", "set_unavailable", "set_error", "state", "dataset"):
            self.__dict__.pop(k, None)

    def __repr__(self):
        return u"<Dataset name='{}' id='{}'>".format(self.name, self.id)

    @property
    def collaborators(self):
        """
        List of collaborators on Dataset.
        """
        self._check_exists()
        return self._api.datasets.get_collaborators(self)

    def add_collaborators(self, *collaborator_ids):
        """
        Add new collaborator(s) to Dataset.

        Args:
            collaborator_ids: List of collaborator IDs to add (Users, Organizations, Teams)
        """
        self._check_exists()
        return self._api.datasets.add_collaborators(self, *collaborator_ids)

    def remove_collaborators(self, *collaborator_ids):
        """
        Remove collaborator(s) from Dataset.

        Args:
            collaborator_ids: List of collaborator IDs to remove (Users)
        """
        self._check_exists()
        return self._api.datasets.remove_collaborators(self, *collaborator_ids)

    @property
    def _get_method(self):
        return self._api.datasets.get

    def as_dict(self):
        return dict(
            name = self.name,
            description = self.description,
            properties = [p.as_dict() for p in self.properties]
        )

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Collections
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class Collection(BaseCollection):
    def __init__(self, name, **kwargs):
        kwargs.pop('package_type', None)
        super(Collection, self).__init__(name, package_type="Collection", **kwargs)

    @property
    def _get_method(self):
        return self._api.packages.get

    def __repr__(self):
        return u"<Collection name='{}' id='{}'>".format(self.name, self.id)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Data Ledger
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class LedgerEntry(BaseNode):
    def __init__(self,
            reference,
            userId,
            organizationId,
            metric,
            value,
            date):

        super(LedgerEntry, self).__init__()
        self.reference = reference
        self.userId = userId
        self.organizationId = organizationId
        self.metric = metric
        self.value = value
        self.date = date

    @classmethod
    def from_dict(self, data):
        return LedgerEntry(data["reference"],
                data["userId"],
                data["organizationId"],
                data["metric"],
                data["value"],
                dateutil.parser.parse(data["date"]))

    def as_dict(self):
        return {
                "reference": self.reference,
                "userId": self.userId,
                "organizationId": self.organizationId,
                "metric": self.metric,
                "value": self.value,
                "date": self.date.replace(microsecond=0).isoformat() + 'Z'
                }




import os
import configparser
import sys
import tempfile

class Settings(object):
    def __init__(self):
        self.profiles = {}

        self._load_defaults()
        self._load_global()
        self._load_profiles()

        try:
            self.use_profile(self.config['global']['default_profile'])
        except:
            self.use_profile('global')

        self._load_eVars()

    def _load_defaults(self):
        # GET/MAKE REQUIRED DIRECTORIES/FILES
        #=============================================
        self.blackfynn_dir = os.environ.get('BLACKFYNN_LOCAL_DIR', os.path.join(os.path.expanduser('~'), '.blackfynn'))
        if not os.path.exists(self.blackfynn_dir):
            os.makedirs(self.blackfynn_dir)
            
        self.cache_dir = os.environ.get('BLACKFYNN_CACHE_LOC', os.path.join(self.blackfynn_dir,'cache'))
        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)

        self.working_dataset_file = os.path.join(tempfile.gettempdir(), 'blackfynn_working_dataset_{}'.format(os.getppid()))
        if os.path.exists(self.working_dataset_file):
            f = open(self.working_dataset_file, 'r+')
            self.working_dataset = f.read()
            f.close()
        else:
            self.working_dataset = None

        # GET CONFIG OBJECT
        #============================================
        self.config_file = os.path.join(self.blackfynn_dir,'config.ini')
        self.config = configparser.ConfigParser()
        self.config.read(self.config_file)

            
        # DEFAULT SETTINGS
        #=============================================
        self.defaults = {    
            # blackfynn api locations
            'api_host'                    : 'https://api.blackfynn.io',
            'streaming_api_host'          : 'https://streaming.blackfynn.io',
            
            # blackfynn API token/secret
            'api_token'                   : None,
            'api_secret'                  : None,
            
            # streaming
            'stream_name'                 : 'prod-stream-blackfynn',
            'stream_aws_region'           : 'us-east-1',
            'stream_max_segment_size'     : 5000,
            
            # all requests
            'max_request_time'            : 120, # two minutes
            'max_request_timeout_retries' : 2,
            
            #io
            'max_upload_workers'          : 10,
            
            # timeseries
            'max_points_per_chunk'        : 10000,
            
            # s3 (amazon/local)
            's3_host'                     : '',
            's3_port'                     : '',
            
            # logging
            'log_level'                   : 'INFO',
            
            # cache
            'cache_index'                 : os.path.join(self.cache_dir, 'index.db'),
            'cache_max_size'              : 2048, #os.environ.get('BLACKFYNN_CACHE_MAX_SIZE', 2048) # in MBs
            'cache_inspect_interval'      : 1000, #os.environ.get('BLACKFYNN_CACHE_INSPECT_EVERY', 1000) # num page writes
            'ts_page_size'                : 3600, #int(os.environ.get('BLACKFYNN_TS_PAGE_SIZE', 3600)) # num points / page
            'use_cache'                   : True, #bool(int(os.environ.get('BLACKFYNN_USE_CACHE', True)))
        }

    def _load_global(self):
        self.default_profile = None
        self.profiles['global'] = self.defaults.copy()
        if 'global' in self.config:
            for key, value in self.config['global'].items():
                if value == 'none'            : self.profiles['global'][key] = None
                elif value.lower() == 'true'  : self.profiles['global'][key] = True
                elif value.lower() == 'false' : self.profiles['global'][key] = False
                elif value.isdigit()          : self.profiles['global'][key] = int(value)
                else                          : self.profiles['global'][key] = str(value)

    def _load_eVars(self):
        eVars_dict = {
            'api_host'               : ('BLACKFYNN_API_LOC', str),
            'streaming_api_host'     : ('BLACKFYNN_STREAMING_API_LOC', str),
            'api_token'              : ('BLACKFYNN_API_TOKEN', str),
            'api_secret'             : ('BLACKFYNN_API_SECRET', str),
            'stream_name'            : ('BLACKFYNN_STREAM_NAME', str),
            'working_dataset'        : ('BLACKFYNN_WORKING_DATASET', str),
            
            'cache_max_size'         : ('BLACKFYNN_CACHE_MAX_SIZE', int),
            'cache_inspect_interval' : ('BLACKFYNN_CACHE_INSPECT_EVERY', int),
            'ts_page_size'           : ('BLACKFYNN_TS_PAGE_SIZE', int),
            'use_cache'              : ('BLACKFYNN_USE_CACHE', lambda x: bool(int(x))),
            'log_level'              : ('BLACKFYNN_LOG_LEVEL', str),
            'default_profile'        : ('BLACKFYNN_PROFILE', str),

            # advanced
            's3_host'                : ('S3_HOST', str),
            's3_port'                : ('S3_PORT', str)
        }

        self.eVars = {}
        values_override = {}
        for key, (evar, typefunc) in eVars_dict.items():
            value = os.environ.get(evar,None)
            if value is not None:
                v = typefunc(value)
                self.eVars[key] = [v, evar]
                values_override[key] = v

        if 'default_profile' in self.eVars:
            try:
                self.use_profile(self.eVars['default_profile'][0])
            except:
                sys.exit("Invalid value '{}' for environment variable {}".format(self.eVars['default_profile'][0],self.eVars['default_profile'][1]))

        self.__dict__.update(values_override)
                
    def _load_profiles(self):
        for name in self.config.sections():
            if name is not 'global':
                self.profiles[name] = self.profiles['global'].copy()
                for key, value in self.config[name].items():
                    if value == 'none'            : self.profiles[name][key] = None
                    elif value.lower() == 'true'  : self.profiles[name][key] = True
                    elif value.lower() == 'false' : self.profiles[name][key] = False
                    elif value.isdigit()          : self.profiles[name][key] = int(value)
                    else                          : self.profiles[name][key] = str(value)

    def use_profile(self, name):
        if name is not None:
            if name not in self.profiles:
                raise Exception('Invaid profile name')
            else:
                self.__dict__.update(self.profiles[name])
                self.active_profile = name
                if name is 'global': self.active_profile = 'none'

    def use_dataset(self, dataset_id):
        self.working_dataset = dataset_id

settings = Settings()



import os
import pytest
import uuid

from blackfynn import Blackfynn

def pytest_addoption(parser):
    parser.addoption("--devserver", default=[], help=("Test against dev server (not local)"))

def pytest_generate_tests(metafunc):
    print metafunc.fixturenames
    use_dev = False
    if 'devserver' in metafunc.fixturenames:
        use_dev = metafunc.config.option.devserver
    metafunc.parametrize("use_dev", [use_dev], scope='session')
    

@pytest.fixture(scope='session')
def client(use_dev):
    """
    Login via API, return client. Login information, by default, will be taken from
    environment variables, so ensure those are set properly before testing. Alternatively,
    to force a particular user, adjust input arguments as necessary.
    """
    bf = Blackfynn(
        api_token='43db6624-68ac-4e60-8ee6-d248b0db0445',
        api_secret='c1a421e3-6f2e-4387-8d3b-e0596bd6a711'
    )
    # get organizations
    orgs = bf.organizations()
    print 'organizations =', orgs
    assert len(orgs) > 0

    # explicitly set context to Blackfyn org
    assert bf.context is not None
    return bf

@pytest.fixture(scope='session')
def client2(use_dev):
    bf = Blackfynn(
        api_token='43db6624-68ac-4e60-8ee6-d248b0db0445',
        api_secret='c1a421e3-6f2e-4387-8d3b-e0596bd6a711'
    )
    # get organizations
    orgs = bf.organizations()
    assert len(orgs) > 0

    # explicitly set context to Blackfyn org
    assert bf.context is not None
    return bf


@pytest.fixture(scope='session')
def superuser_client(use_dev):
    """
    Client using super-admin permissions
    """
    bf = Blackfynn(
        api_token='c535152c-0564-4865-8fa0-addebb07e020',
        api_secret='89d1a675-a876-4e16-8dd7-b54f58ff3c8e'
    )
    assert bf.profile.is_super_admin
    return bf


@pytest.fixture(scope='session')
def dataset(use_dev, client, superuser_client):
    """
    Test Dataset to be used by other tests.
    """

    # collection of all datasets
    n_ds = len(client.datasets())

    # create test dataset
    ds = client.create_dataset("test dataset {}".format(uuid.uuid4()))
    assert ds.exists
    assert len(client.datasets()) == n_ds + 1

    # surface test dataset to other functions
    yield ds

    # remove
    superuser_client._api.datasets.delete(ds)

    assert len(client.datasets()) == n_ds
    assert not ds.exists
    assert not hasattr(ds, 'parent')

@pytest.fixture(scope='session')
def test_organization(client):
    return filter(lambda o: o.name == 'Test Organization', client.organizations())[0]

def test_login(client):
    email = os.environ.get('BLACKFYNN_USER')
    profile = client.profile
    print "profile = ", profile
    assert profile['email'] == email




'''
usage:
  bf [options] init <name>

global options:
  -h --help                 Show help
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt
import sys

from cli_utils import get_client, set_working_dataset, print_datasets
from blackfynn import settings

def main():
    args = docopt(__doc__)

    bf = get_client()

    name = args['<name>']
    dataset = bf.create_dataset(name)
    set_working_dataset(dataset)

    print_datasets(bf)




import time
import datetime

# blackfynn
from blackfynn import Blackfynn, TimeSeries, TimeSeriesChannel
from blackfynn.streaming import TimeSeriesStream
from blackfynn.utils import generate_dataframe


def test_stream_upload(use_dev, client, dataset):
    # can only run against dev server :-(
    if not use_dev: return

    # generate data
    freq = 100
    df = generate_dataframe(minutes=1, freq=100)

    # create timeseries
    ts = TimeSeries('My Test TimeSeries')
    dataset.add(ts)
    assert ts.exists
    print ts

    # create channels
    channels = [TimeSeriesChannel(c, rate=freq) for c in df.columns]
    ts.add_channels(*channels)
    for ch in channels:
        assert ch.exists
        assert ch.rate == freq
        assert ch.name in df.columns
    print "channels =", ts.channels

    # stream data up
    ts.stream_data(df)

    # wait a bit -- let server write data
    time.sleep(10)

    # retrieve data
    for chunk in ts.get_data_iter():
        assert len(chunk) > 0
        assert chunk.columns == df.columns
    assert chunk.index[-1] == df.index[-1]

    # make sure channel times were updated
    for ch in ts.channels:
        assert ch.start_datetime == df.index[0]
        assert ch.end_datetime == df.index[-1]



# -*- coding: utf-8 -*-

import re
import math
import datetime
import numpy as np
import pandas as pd
from types import NoneType
from itertools import islice, count
from concurrent.futures import ThreadPoolExecutor

# blackfynn
from blackfynn.api.base import APIBase
from blackfynn.streaming import TimeSeriesStream
from blackfynn.utils import (
    usecs_to_datetime, usecs_since_epoch, infer_epoch, log
)
from blackfynn.models import (
    File, TimeSeries,TimeSeriesChannel, TimeSeriesAnnotation, 
    get_package_class, TimeSeriesAnnotation, TimeSeriesAnnotationLayer
)
from blackfynn import settings
from blackfynn.cache import get_cache

cache = None
page_size = settings.ts_page_size

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Helpers
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def parse_timedelta(time):
    """
    Returns microseconds of time expression, where time can be of the forms:
     - string:  e.g. '1s', '5m', '3h'
     - delta:   datetime.timedelta object
    """
    if isinstance(time, basestring):
        # parse string into timedelta
        regex = re.compile(r'((?P<hours>\d*\.*\d+?)hr)?((?P<minutes>\d*\.*\d+?)m)?((?P<seconds>\d*\.*\d+?)s)?')
        parts = regex.match(time)
        if not parts:
            return
        parts = parts.groupdict()
        time_params = {}
        for (name, param) in parts.iteritems():
            if param:
                time_params[name] = float(param)
        time = datetime.timedelta(**time_params)

    if isinstance(time, datetime.timedelta):
        # return microseconds
        return long(time.total_seconds()*1e6)

    elif isinstance(time, (long, int, float)):
        # assume already in microseconds
        return time

vec_usecs_to_datetime = np.vectorize(usecs_to_datetime)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# TimeSeries Request
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class ChannelPage(object):
    def __init__(self, channel, page, use_cache=True):
        self.channel   = channel
        self.page      = long(page)
        self.use_cache = use_cache
        
        global cache, page_size
        if self.use_cache and cache is None:
            cache = get_cache(start_compaction=True)
            page_size = cache.page_size
        
        # fixed page -- determined from epoch(0)
        pg_delta = channel._page_delta(page_size)
        self.start = long(self.page  * pg_delta)
        self.stop  = long(self.start + pg_delta)
        self.cache_exists = False

        # current future/response
        self.future = None
        self.data   = None

    def request(self, api):
        # check if page is cached
        if self.use_cache and cache.check_page(self.channel, self.page):
            # we (should) have cache, skip API request
            self.cache_exists = True
            return

        # make request: not using cache
        args = dict(
            # Note: uses streaming server
            host     = api._streaming_host,
            endpoint = '/ts/retrieve',
            base     = '',
            async    = True,
            params   = dict(
                channel = self.channel.id,
                limit   = '', # required by API
                session = api.headers.get('X-SESSION-ID'),
                start   = self.start,
                end     = self.stop)
        )
        self.future = api._get(**args)

    def get(self, api, update_cache_if_exists=False):
        if self.future is not None:
            # we're handling an API request/response
            self.data = self._get_response(api)

            # set cache!
            if self.use_cache and (not self.cache_exists or update_cache_if_exists):
                cache.set_page_data(self.channel, self.page, self.data, update=update_cache_if_exists)

        elif self.data is not None:
            # we've already got the result
            pass

        elif self.use_cache and self.cache_exists:
            # use existing cache entry
            self.data = cache.get_page_data(self.channel, self.page)

            if self.data is None:
                # cache may have disappeared, let's make API call
                self.use_cache = False
                self.cache_exists = False
                self.request(api)
                self.data = self.get(api, update_cache_if_exists=True)

        return self.data

    def _get_response(self, api, datetime_index=True):
        # handle API response, return data series
        resp  = api._get_response(self.future)
        self.future = None

        # handle data response
        times = np.array(resp['times'])
        data  = np.array(resp['data'])
        # fix -- sometimes API responds out-of-order
        order = np.argsort(times)
        times = times[order]
        data  = data[order]

        if datetime_index and len(times)>0:
            times = vec_usecs_to_datetime(times)

        # return pandas series
        return pd.Series( data=data, index=times, name=resp['channel'])


class ChannelIterator(object):
    """
    We make requests to API/cache using some fixed page-size, but the 
    user typically wants data results in some specified "chunk size".
    This accumulates the data pages in order to serve the data back
    in the specified chunk size.
    """
    def __init__(self, channel, start, stop, chunk_time, api, use_cache=True):
        self.channel    = channel
        self.start      = start
        self.stop       = stop
        self.use_cache  = use_cache
        self.api        = api

        # page delta (usecs) for channel
        self.page_delta = channel._page_delta(page_size)

        # page iteration
        self.page_start = long(math.floor(self.start/(1.0*self.page_delta)))
        self.page_end   = long(math.ceil(self.stop /(1.0*self.page_delta)))

        # chunk iteration
        self.chunk_per_page = chunk_time is None
        # chunk over specfied time
        if not self.chunk_per_page:
            self.chunk_time = long(chunk_time) # in usecs
            self.chunk_size = long(channel.rate * self.chunk_time/1.0e6)
        self.chunk      = None

    def get_chunks(self):
        # page size may be more/less than requested data

        self.chunk = pd.Series()
        pages = iter(np.arange(self.page_start, self.page_end))
        page = None
        while True:
            if self.chunk_per_page or \
               (page is None and len(self.chunk) < self.chunk_size):
                # get next page
                try: p = next(pages)
                except: break
                page = ChannelPage(
                        channel   = self.channel,
                        page      = p,
                        use_cache = self.use_cache)
                page.request(self.api)
                data = page.get(self.api)
                # no more data
                if data is None: break
                # grow data series
                i_start = None
                i_stop  = None
                if page.start < self.start:
                    i_start = usecs_to_datetime(self.start)
                if page.stop > self.stop:
                    i_stop = usecs_to_datetime(self.stop-1)
                else:
                    # more pages needed - reset
                    page = None
                data_slice = data.ix[i_start:i_stop]
                if self.chunk_per_page:
                    yield data_slice
                else:
                    self.chunk = self.chunk.append(data_slice)
            else:
                # return full chunk
                if not self.chunk_per_page:
                    yield self._get_chunk()

        # return remaining chunk
        if not self.chunk_per_page:
            yield self._get_chunk()

    def _get_chunk(self):
        d = self.chunk.ix[:self.chunk_size]
        # leave remainder
        self.chunk = self.chunk.ix[self.chunk_size:]
        return d if len(d) else None

    def __repr__(self):
        return "<ChannelIterator channel='{}' range=({},{})>".format(
                    self.channel.id, self.start, self.stop)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Time Series API
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class TimeSeriesAPI(APIBase):
    base_uri = "/timeseries"
    name = 'timeseries'

    # ~~~~~~~~~~~~~~~~~~~
    # Channels
    # ~~~~~~~~~~~~~~~~~~~

    def create_channel(self, ts, channel):
        """
        Adds a channel to a timeseries package on the platform.
        """
        if channel.exists:
            return channel
        ts_id = self._get_id(ts)
        resp = self._post( self._uri('/{id}/channels', id=ts_id), json=channel.as_dict())

        ch = TimeSeriesChannel.from_dict(resp, api=self.session)
        ch._pkg = ts.id
        return ch

    def get_channels(self, ts):
        """
        Returns a set of channels for a timeseries package.
        """
        ts_id = self._get_id(ts)
        resp = self._get( self._uri('/{id}/channels', id=ts_id) )

        chs = [TimeSeriesChannel.from_dict(r, api=self.session) for r in resp]
        for ch in chs:
            ch._pkg = ts_id
        return chs

    def get_channel(self, pkg, channel):
        """
        Returns a channel object from the platform.
        """
        pkg_id = self._get_id(pkg)
        channel_id = self._get_id(channel)
        
        path = self._uri('/{pkg_id}/channels/{id}', pkg_id=pkg_id, id=channel_id)
        resp = self._get(path)

        ch = TimeSeriesChannel.from_dict(resp, api=self.session)
        ch._pkg = pkg_id
        return ch

    def update_channel(self, channel):
        """
        Updates a channel on the platform.

        Note: must be super-admin.
        """
        ch_id = self._get_id(channel)
        pkg_id = self._get_id(channel._pkg)
        path = self._uri('/{pkg_id}/channels/{id}', pkg_id=pkg_id, id=ch_id)

        resp = self._put(path , json=channel.as_dict())

        ch = TimeSeriesChannel.from_dict(resp, api=self.session)
        ch._pkg = pkg_id
        return ch

    def update_channel_properties(self, channel):
        """
        Updates a channel's properties on the platform.

        Note: must be super-admin.
        """
        ch_id = self._get_id(channel)
        pkg_id = self._get_id(channel._pkg)
        path = self._uri('/{pkg_id}/channels/{id}/properties', pkg_id=pkg_id, id=ch_id)

        resp = self._put(path , json=[m.as_dict() for m in channel.properties])

        ch = TimeSeriesChannel.from_dict(resp, api=self.session)
        ch._pkg = pkg_id
        return ch

    def delete_channel(self, channel):
        """
        Deletes a timeseries channel on the platform.
        """
        
        ch_id = self._get_id(channel)
        pkg_id = self._get_id(channel._pkg)
        path = self._uri('/{pkg_id}/channels/{id}', pkg_id=pkg_id, id=ch_id)

        return self._del(path)

    def get_streaming_credentials(self, ts):
        """
        Get the streaming credentials for the given time series package.

        NOTE: this does not currently returns credentials but is still needed
        to make sure the package is registered in the system as a streaming
        time series package.
        """

        pkg_id = self._get_id(ts)

        return self._get(self._uri('/{pkg_id}/streaming_credentials', pkg_id=pkg_id))

    # ~~~~~~~~~~~~~~~~~~~
    # Data
    # ~~~~~~~~~~~~~~~~~~~

    def get_ts_data_iter(self, ts, start, end, channels, chunk_size, 
                         use_cache,length=None):
        """
        Iterator will be constructed based over timespan (start,end) or (start, start+seconds)

        Both :chunk_size and :length should be described using strings, e.g.
          5 second  = '5s'
          3 minutes = '3m'
          1 hour    = '1h'
        otherwise microseconds assumed.
        """
        if isinstance(ts, basestring):
            # assumed to be package ID
            ts = self.session.core.get(ts)

        #CHANNELS
        ts_channels = ts.channels

        #no channels specified
        if channels is None:
            channels = ts.channels
        #1 channel specified as TSC object
        elif isinstance(channels,TimeSeriesChannel):
            channels = [channels]
        #1 channel specified and channel id
        elif isinstance(channels,basestring):
            channels = [ch for ch in ts.channels if ch.id==channels]
        #list of channel ids OR ts channels
        else:
            all_ch = []
            for chan in channels:
                if isinstance(chan,basestring):
                    all_ch.extend([ch for ch in ts_channels if ch.id==chan])
                else:
                    all_ch.extend([ch for ch in ts_channels if ch==chan])
            channels = all_ch

        # determine start (usecs)
        the_start = ts.start if start is None else infer_epoch(start)

        # chunk 
        if chunk_size is not None and isinstance(chunk_size, basestring):
            chunk_size = parse_timedelta(chunk_size)

        # determine end
        if length is not None:
            if isinstance(length, basestring):
                length_usec = parse_timedelta(length)
            else:
                length_usec = length
            the_end = the_start + length_usec

        elif end is not None:
            the_end = infer_epoch(end)
        else:
            the_end = ts.end

        # logical check
        if the_end < the_start:
            raise Exception("End time cannot be before start time - unless you're magic.")

        # loop through chunks
        the_start = long(the_start)
        the_end = long(the_end)

        channel_chunks = [
            ChannelIterator(ch, the_start, the_end, chunk_size,
                            api=self.session, use_cache=use_cache).get_chunks()
            for ch in channels
        ]

        while True:
            # get chunk for all channels
            values = [next(i, None) for i in channel_chunks]
            # no more results?
            if not [1 for v in values if v is not None]:
                break
            # make dataframe 
            data_map = {c.name: v for c,v in zip(channels,values) if v is not None}
            yield pd.DataFrame.from_dict(data_map) 

    def get_ts_data(self, ts, start, end, length, channels, use_cache):
        """
        Retrieve data. Must specify end-time or length.
        """
        ts_iter = self.get_ts_data_iter(ts=ts, start=start, end=end, channels=channels,
                                         chunk_size=None, use_cache=use_cache, length=length)
        df = pd.DataFrame()
        for tmp_df in ts_iter:
            df = df.append(tmp_df)
        return df

    def stream_data(self, ts, dataframe):
        """
        Stream timeseries data
        """
        stream = TimeSeriesStream(ts)
        return stream.send_data(dataframe)

    def stream_channel_data(self, channel, series):
        """
        Stream channel data
        """
        raise NotImplementedError
        

    # ~~~~~~~~~~~~~~~~~~~
    # Annotation Layers 
    # ~~~~~~~~~~~~~~~~~~~

    def create_annotation_layer(self, ts, layer, description):

        if isinstance(layer,TimeSeriesAnnotationLayer):
            data = layer.as_dict()
        elif isinstance(layer,basestring):
            data = {
                'name' : layer,
                'description' : description
            }
        else:
            raise Exception("Layer must be TimeSeriesAnnotationLayer object or name of new layer")

        existing_layer = [i for i in ts.layers if data['name'] == i.name]
        if existing_layer:
            print 'Returning existing layer {}'.format(existing_layer)
            return existing_layer[0]
        else:
            ts_id = self._get_id(ts)
            path = self._uri('/{id}/layers', id=ts_id)
            resp = self._post(path, json=data)
            tmp_layer = TimeSeriesAnnotationLayer.from_dict(resp, api=self.session)
            if isinstance(layer,TimeSeriesAnnotationLayer):
                layer.__dict__.update(tmp_layer.__dict__)
            return tmp_layer

    def get_annotation_layer(self, ts, layer): 
        ts_id = self._get_id(ts) 
        layer_id = self._get_id(layer)
        path = self._uri('/{id}/layers/{layer_id}',id=ts_id,layer_id=str(layer_id))
        resp = self._get(path)
        return TimeSeriesAnnotationLayer.from_dict(resp, api=self.session)

    def get_annotation_layers(self, ts):
        ts_id = self._get_id(ts)
        resp = self._get(self._uri('/{id}/layers', id=ts_id))
        return [TimeSeriesAnnotationLayer.from_dict(x, api=self.session) for x in resp["results"]]

    def update_annotation_layer(self, ts, layer):
        #return all layers
        ts_id = self._get_id(ts)
        layer_id = self._get_id(layer)
        path = self._uri('/{id}/layers/{layer_id}', id=ts_id, layer_id=layer_id)
        resp = self._put(path, json=layer.as_dict())
        return TimeSeriesAnnotationLayer.from_dict(resp, api=self.session)

    def delete_annotation_layer(self, layer):
        ts_id = layer.time_series_id

        path = self._uri('/{id}/layers/{layer_id}',id = ts_id,  layer_id =layer.id)
        if self._del(path):
            layer.id = None
            return True
        else:
            return False

    # ~~~~~~~~~~~~~~~~~~~
    # Annotations
    # ~~~~~~~~~~~~~~~~~~~

    def delete_annotation(self, annot):
        """
        Deletes a single annotation
        """
        path = self._uri('/{ts_id}/layers/{layer_id}/annotations/{annot_id}',
                    ts_id = annot.time_series_id,
                    layer_id = annot.layer_id,
                    annot_id = annot.id)
        if self._del(path):
            annot.id = None
            return True
        else:
            return False

    def create_annotations(self,layer, annotations):

        all_annotations = []

        if not isinstance(annotations,list):
            annotations = [annotations]

        for annot in annotations:
            tmp = self.create_annotation(layer=layer,annotation=annot)                
            all_annotations.append(tmp)

        #if adding single annotation, return annotation object, else return list
        if len(all_annotations) == 1:
            all_annotations = all_annotations[0]

        return all_annotations

    def create_annotation(self, layer, annotation, **kwargs):
        """
        Creates annotation for some timeseries package on the platform.
        """
        if isinstance(annotation,TimeSeriesAnnotation):
            data = annotation.as_dict()
        elif all(x in kwargs for x in ['start','end']):
            start_time = infer_epoch(kwargs['start'])
            end_time = infer_epoch(kwargs['end'])
            data = {
                'name':'',
                'label':annotation,
                'start':long(start_time),
                'end':long(end_time),
            }
            if kwargs['channel_ids']:
                channel_ids = kwargs['channel_ids']
                if isinstance(channel_ids,basestring):
                    channel_ids = [channel_ids]
                data['channelIds']=channel_ids
            else:
                ts = layer._api.core.get(layer.time_series_id)
                data['channelIds']=[x.id for x in ts.channels]
            if 'description' in annotation:
                data['description']=kwargs['description']
            else:
                data['description']=None
        else:
            raise Exception("Must provide TimeSeriesAnnotation object or 'annotation','start','end' at minimum")

        data['time_series_id'] = layer.time_series_id
        data['layer_id'] = layer.id

        path = self._uri('/{ts_id}/layers/{layer_id}/annotations',
                    ts_id=layer.time_series_id, layer_id=layer.id)
        resp = self._post(path, json=data)
        tmp = TimeSeriesAnnotation.from_dict(resp, api=self.session)

        if isinstance(annotation,TimeSeriesAnnotation):
            annotation.__dict__.update(tmp.__dict__)

        return tmp

    def update_annotation(self, ts, layer, annot):
        """
        Update annotation on the platform.
        """
        path = self._uri('/{ts_id}/layers/{layer_id}/annotations/{annot_id}',
                    ts_id = self._get_id(ts),
                    layer_id = self._get_id(layer),
                    annot_id = self._get_id(annot))
        resp = self._put(path, json=annot.as_dict())
        return TimeSeriesAnnotation.from_dict(resp, api=self.session)

    def get_annotation(self, ts, layer, annot):
        """
        Returns a timeseries annotation
        """
        path = self._uri('/{ts_id}/layers/{layer_id}/annotations/{annot_id}',
                    ts_id = self._get_id(ts),
                    layer_id = self._get_id(layer),
                    annot_id = self._get_id(annot))
        resp = self._get(path)
        return TimeSeriesAnnotation.from_dict(resp["annotation"], api=self.session)

    def iter_annotations(self, ts, layer, window_size=10, channels=None):
        # window_size is seconds
        if not isinstance(ts, TimeSeries):
            raise Exception("Argument 'ts' must be TimeSeries.")

        if channels is None:
            # use all channels
            channels = ts.channels
        else:
            # make sure specified channels are in current timeseries
            ts_ch_ids = [ch.id for ch in ts.channels]
            for ch in channels:
                if ch in ts_channels:
                    continue
                raise Exception("Channel '{ch}' not found in TimeSeries '{ts}'".format(
                        ts = ts.id,
                        ch = self._get_id(ch)))

        # paginate annotations
        start_time, end_time = ts.limits()
        num_windows = (end_time-start_time)/(window_size*1e6)
        for i in range(int(np.ceil(num_windows))):
            win_start = start_time + i* (window_size*1e6)
            win_end = win_start + window_size*1e6
            if win_end > end_time:
                win_end = end_time
            annotations = self.query_annotations(ts=ts,layer=layer,start=win_start,end=win_end, channels=channels)
            yield annotations

    def get_annotations(self, ts, layer, channels=None):
        """
        Returns all annotations for a given layer
        """
        start, end = ts.limits()
        return self.query_annotations(ts=ts, layer=layer, start=start, end=end, channels=channels, limit=0, offset=0)


    def query_annotations(self, ts, layer, start=None, end=None, channels=None, limit=None, offset=0):
        """
        Retrieves timeseries annotations for a particular range  on array of channels.
        """
        if channels is None:
            ch_list = [] #empty uses all channels
        else:
            ch_list = [self._get_id(x) for x in channels]

        ts_start, ts_end = ts.limits()
        if start is None:
            start = ts_start
        elif isinstance(start, datetime.datetime):
            start = usecs_since_epoch(start)

        if end is None:
            end = ts_end
        elif isinstance(end, datetime.datetime):
            end = usecs_since_epoch(end)

        params = {
            'start': long(start),
            'end': long(end),
            'channelIds': ch_list,
            'layerName': layer.name,
            'limit': limit,
            'offset': offset
        }
        path = self._uri('/{ts_id}/layers/{layer_id}/annotations',
                    ts_id = ts.id,layer_id = layer.id)

        resp = self._get(path, params=params)

        return [TimeSeriesAnnotation.from_dict(x, api=self.session) for x in resp['annotations']['results']]

    def annotation_counts(self, ts, layer, start, end, period, channels=None):
        """
        Retrives annotation counts for a given ts, channel, start, end, and/or layer
        """
        if channels is None:
            ch_list = [] #empty uses all channels
        else:
            ch_list = [self._get_id(x) for x in channels]

        if isinstance(start, datetime.datetime):
            start = usecs_since_epoch(start)
        if isinstance(end, datetime.datetime):
            end = usecs_since_epoch(end)

        period = parse_timedelta(period)

        params = {
            'start': long(start),
            'end': long(end),
            'channelIds': ch_list,
            'period': period,
            'layer': layer.name
        }

        path = self._uri('/{ts_id}/annotations',
                    ts_id = self._get_id(ts))

        resp = self._get(path, params=params)
        return resp

    def process_annotation_file(self,ts,file_path):
        """
        Processes the .bfannot file at file_path and adds to timeseries package
        """
        if not file_path.lower().endswith(('.bfannot')):
            raise Exception("Annotation file format not currently supported. Supported annotations types: .bfannot")
        try:
            df = pd.read_csv(file_path)
            df = df.where((pd.notnull(df)), None)
            channels = ts.channels
            if (df['version'][0] == 1.0): #version number
                layers = df['layer_name'].unique()
                for l in layers:
                    layer = []
                    annots = df.loc[df['layer_name'] == l]

                    #create or find existing layer
                    layer = ts.add_layer(layer=l, description=annots['layer_description'].iloc[0])

                    annotations = []
                    for index, row in annots.iterrows():
                        if pd.isnull(row['channel_names']):
                            channel_ids = [x.id for x in channels]
                        else:
                            channel_names = row['channel_names'].split(';')
                            channel_ids = [x.id for x in channels if x.name in channel_names]

                        layer.insert_annotation(annotation=row['annotation_label'],channel_ids=channel_ids,start=row['start_uutc'],end=row['end_uutc'],description=row['annotation_description'])

                    print 'Added annotations to layer {} , pkg: {}'.format(layer,ts)
            else:
                raise Exception('BF version {} not found or not supported'.format(df['version'][0]))

        except Exception, error:
            raise Exception("Error adding annotation file {}, {}".format(file_path, error))

    def write_annotation_file(self,ts,file_path,layer_names):
        """
        Writes all layers in ts to .bfannot (v1.0) file

        """
        layers = ts.layers
        if not layers:
            raise Exception('Timeseries has no existing layers')

        if not file_path.lower().endswith(('.bfannot')):
            file_path+='.bfannot'

        headers = ['version',
                    'package_type',
                    'layer_name',
                    'layer_description',
                    'annotation_label',
                    'start_uutc',
                    'end_uutc',
                    'channel_names',
                    'annotation_description']
        out = pd.DataFrame(columns=headers)

        if layer_names:
            if not isinstance(layer_names,list):
                layer_names = [layer_names]
            new_layers = [l for l in layers if l.name in layer_names]
            layers = new_layers

        to_write = []
        for l in layers:
            annot = l.annotations()
            for a in annot:
                channels = [ch.name for ch in ts.channels if ch.id in a.channel_ids]
                channel_names = ";".join(channels)
                tmp = {
                    'layer_name' : l.name,
                    'layer_description': l.description,
                    'annotation_label' : a.label,
                    'start_uutc' : long(a.start),
                    'end_uutc' : long(a.end),
                    'channel_names' : channel_names,
                    'annotation_description' : a.description
                }
                to_write.append(tmp)
        out = out.append(to_write)

        # Add version number and package type
        out.loc[0,('version')] = 1.0
        out.loc[0,('package_type')] = ts.type

        out.to_csv(file_path,index=False)

    # ~~~~~~~~~~~~~~~~~~~
    # Helpers
    # ~~~~~~~~~~~~~~~~~~~


    def _annotation_query_params(ts, start, end, period, layer, channels):
        # parse channel input
        channels = self._channel_list(ts, channels)

        # channel IDs
        ch_ids = [x.id for x in channels]

        params = {
            'start': usecs_since_epoch(start),
            'end': usecs_since_epoch(end),
            'channels': ch_ids,
            'period': period
        }

        if layer is not None:
            params['layer'] = layer
        return params

    def _channel_list(self, ts, channels):
        """ 
        Get list of channel objects provided flexible input values
        """ 
        ts_id = self._get_id(ts)

        if channels is None:
            # if channel(s) not specified, grab all for package
            channels = self.session.timeseries.get_channels(ts_id)

        # check if list
        if not hasattr(channels, '__iter__'):
            # they specified a single object
            channels = [channels]

        # check type of items in list
        for ch in channels:
            if isinstance(ch, TimeSeriesChannel):
                # Channel looks good
                continue
            if isinstance(ch, basestring):
                # Assume channel ID, get object
                ch = self.session.get(ch)
            else:
                raise Exception('Expecting TimeSeries instance or ID') 

        return channels





'''
usage:
  bf where [options] <item>

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt
from cli_utils import print_path_tree, get_client, get_item
        
def main():
    args = docopt(__doc__)

    bf = get_client()

    item = get_item(args['<item>'], bf)
    print_path_tree(bf, [item])



'''
usage:
  bf [options] ( datasets | ds )

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt
import sys

from cli_utils import get_client, print_datasets

def main():
    args = docopt(__doc__)

    print_datasets(get_client())



'''
usage:
  bf upload [options] [--to=destination] <file>...

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt
from blackfynn import settings
from cli_utils import recursively_upload, get_client
import os

def main():
    args = docopt(__doc__)

    bf = get_client()

    files = args['<file>']

    if args['--to']:
        collection = bf.get(args['destination'])
        recursively_upload(bf, collection, files)

    elif settings.working_dataset:
        dataset = bf.get_dataset(settings.working_dataset)
        recursively_upload(bf, dataset, files)



'''
usage:
  bf [options] ( organizations | orgs )

global options:
  -h --help                 Show help
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt
import sys

from cli_utils import get_client

def main():
    args = docopt(__doc__)

    bf = get_client()

    for o in bf.organizations():
        print "  {} (id: {})".format(o.name, o.id)



'''
usage:
  bf delete [options] <item>

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt
from blackfynn import Dataset
from blackfynn.models import BaseNode

from cli_utils import get_client, get_item

def main():
    args = docopt(__doc__)

    bf = get_client()

    try:
        item = get_item((args['<item>'], bf)

        if not (isinstance(item, Collection) or isinstance(item, DataPackage)):
            raise Exception("only data packages and collections may be deleted.")

        bf.delete(item)
    except Exception, e:
        exit("  failed to delete {}: {}".format(args['<item>']), e)



'''
usage:
  bf status [options]

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''
from docopt import docopt

from blackfynn import settings

from cli_utils import get_client

def main():
    args = docopt(__doc__)

    print('Active profile:\n  \033[32m{}\033[0m\n'.format(settings.active_profile))

    if len(settings.eVars.keys()) > 0:
        key_len = 0
        value_len = 0
        for key, (value,_) in settings.eVars.items():
            valstr = '{}'.format(value)
            key_len = max(key_len,len(key))
            value_len = max(value_len,len(valstr))
        
        print('Environment variables:')
        print('  \033[4m{:{key_len}}    {:{value_len}}    {}'.format('Key','Value','Environment Variable\033[0m',key_len=key_len,value_len=value_len))
        for key, (value,evar) in sorted(settings.eVars.items()):
           valstr = '{}'.format(value)
           print('  {:{key_len}}    {:{value_len}}    {}'.format(key,valstr,evar,key_len=key_len,value_len=value_len))
        print
           
    bf = get_client()

    if settings.working_dataset:
        try:
            working_dataset = bf.get_dataset(settings.working_dataset)
            working_dataset_status = "{} (id: {})".format(working_dataset.name, working_dataset.id)
        except:
            working_dataset_status = 'Not set.'
    else:
        working_dataset_status = 'Not set.'

    print "Blackfynn environment:"
    print "  User               : {}".format(bf.profile.email)
    print "  Organization       : {} (id: {})".format(bf.context.name, bf.context.id)
    print "  Dataset            : {}".format(working_dataset_status)
    print "  API Location       : {}".format(bf.host)
    print "  Streaming API      : {}".format(settings.streaming_api_host)




import urllib
from blackfynn.models import get_package_class

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Base class
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class APIBase(object):
    base_uri = ''
    name = ''

    def __init__(self, session):
        """
        Base class to be used by all API components.
        """
        # api session
        self.session = session

    def _get_id(self, thing):
        """
        Get ID for object. Assumes string is already ID.
        """
        if isinstance(thing, (basestring, int, long)):
            return thing
        elif thing is None:
            return None
        else:
            return thing.id

    def _get_package_from_data(self, data):
        # parse json
        cls = get_package_class(data)
        pkg = cls.from_dict(data, api=self.session)

        return pkg

    def _uri(self, url_str, **kwvars):
        vals = {k:urllib.quote(str(var)) for k,var in kwvars.items()}
        return url_str.format(**vals)

    def _get(self, endpoint, async=False, base=None, *args, **kwargs):
        base = self.base_uri if base is None else base
        return self.session._call('get', endpoint, base=base, async=async, *args, **kwargs)

    def _post(self, endpoint, async=False, base=None, *args, **kwargs):
        base = self.base_uri if base is None else base
        return self.session._call('post', endpoint, base=base, async=async, *args, **kwargs)

    def _put(self, endpoint, async=False, base=None, *args, **kwargs):
        base = self.base_uri if base is None else base
        return self.session._call('put', endpoint, base=base, async=async, *args, **kwargs)

    def _del(self, endpoint, async=False, base=None, *args, **kwargs):
        base = self.base_uri if base is None else base
        return self.session._call('delete', endpoint, base=base, async=async, *args, **kwargs)

    def _get_response(self, req):
        return self.session._get_response(req)



'''
usage:
  bf use [options] <dataset>

global options:
  -h --help                 Show help
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt

from cli_utils import get_client, set_working_dataset
from blackfynn import settings

def main():
    args = docopt(__doc__)

    bf = get_client()
    dataset_id_or_name = args['<dataset>']

    try:
        dataset = bf.get_dataset(dataset_id_or_name)

        set_working_dataset(dataset)
    except Exception, e:
        exit(e)



import time
import boto3
import base64
import random
import hashlib
import datetime
import numpy as np
import pandas as pd

# blackfynn
from blackfynn import settings
from blackfynn.streaming.segment_pb2 import IngestSegment
from blackfynn.utils import usecs_since_epoch

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Time Series Stream (upload)
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class TimeSeriesStream():

    def __init__(self, ts):
        self.name = settings.stream_name
        self.max_segment_size = settings.stream_max_segment_size

        self.conn = boto3.client('kinesis', region_name=settings.stream_aws_region)

        # reference time-series
        self.ts = ts
        # cache channels
        self._channels = ts.channels

        self.registered = False
    
    @property
    def status(self):
        r = conn.describe_stream(self.name)
        description = r.get('StreamDescription')
        return description.get('StreamStatus')

    def _channel_by_name(self, name):
        matches = filter(lambda c: c.name==name, self._channels)
        if len(matches) > 1:
            raise Exception("Too many channels match name '{}'".format(name))
        if len(matches) == 0:
            raise Exception("No channels match name '{}'".format(name))
        return matches[0]

    def _channel_by_id(self, id):
        matches = filter(lambda c: c.id==id, self._channels)
        if len(matches) == 0:
            raise Exception("No channels match ID '{}'".format(id))
        return matches[0]

    def wait_until_ready(self, timeout=120):
        """
        Waits until Kinesis stream is ACTIVE
        """
        timeout = 20
        start = datetime.datetime.now()
        while self.status != 'ACTIVE':
            # sleep for a second
            time.sleep(1.5) 
            now = datetime.datetime.now()
            if (now-start).total_seconds() > timeout:
                raise Exception("Timeout waiting for stream connection.")

    def _make_ingest_segment(self, start_time,channel_id,period,values):
        segment = IngestSegment()
        if isinstance(start_time, datetime.datetime):
            segment.startTime = usecs_since_epoch(start_time)
        else:
            segment.startTime = start_time
        segment.channelId = channel_id
        segment.samplePeriod = period
        segment.data.extend(values)
        return segment

    def _send_segment(self, seg):
        """
        Send a single segment (protobuf) of data to streaming server
        """
        msg = base64.b64encode(seg.SerializeToString())
        pkey = hashlib.sha256(msg).hexdigest()
        resp = self.conn.put_record(
            StreamName=self.name,
            Data=msg,
            PartitionKey=pkey
        )
        return resp

    def _send_contiguous(self, channel, data, period):
        """
        Provided channel series values, send segments in chunks

        channel:  Blackfynn TimeSeriesChannel
        data:     Pandas Series
        period:   Sample period of data

        """
        for offset in np.arange(0, len(data), self.max_segment_size):
            # divide data into chunks
            chunk = data[offset:offset+self.max_segment_size]
            # make data segment
            seg = self._make_ingest_segment(
                        start_time=chunk.index[0],
                        channel_id=channel.id,
                        period=period,
                        values=chunk.values)
            # send data segment
            resp = self._send_segment(seg)

            if 'ResponseMetadata' not in resp:
                raise Exception("Incorrect response from Kinesis.")
            if resp['ResponseMetadata']['HTTPStatusCode'] != 200:
                raise Exception("Received non-200 response code from Kinesis.")

        # NOTE: this should be done by the streaming consumer (server),
        #       but until then, we'll update channel start/end time
        start = usecs_since_epoch(data.index[0])
        end = usecs_since_epoch(data.index[-1])
        if end > channel.end:
            channel.end = end
        if start < channel.start or channel.start == 0:
            channel.start = start
        channel.update()

    def send_data(self, df):
        """
        Provided a Pandas DataFrame, send streaming data to server. Data 
        is streamed in contiguous chunks. 

        1) Channels *must* already exist on platform.

        2) For each channel, the sample rate *must* match the rate of 
           the channel on the platform. 

        3) df.columns should reflect channels of timeseries packages, where
           df.values are data points. Columns can be channel names or IDs.

        """

        # sanity checks
        if not isinstance(df, pd.DataFrame):
            raise Exception("argument df must be Pandas DataFrame")

        # iterate over channels in data
        for col in df.columns:
            try:
                # match column to channel name
                channel = self._channel_by_name(col)
            except:
                # match column to channel ID
                channel = self._channel_by_id(col)

            self.send_channel_data(channel=channel, series=df[col])

    def send_channel_data(self, channel, series):

        if not isinstance(series, pd.Series):
            raise Exception("Series must be a Pandas Series object.")

        if self.registered == False:
            self.ts.streaming_credentials()
            self.registered = True

        # period in microseconds
        period = 1e6/channel.rate 

        # find timestamp differences
        ind = series.index.values
        ts_steps = (ind[1:] - ind[:-1]).astype('timedelta64[us]').astype(int)

        # find gaps
        gap_fuzz = period*0.75
        gaps = ts_steps > period + gap_fuzz
        gaps_ind = np.where(gaps)[0]
        gap_pairs = zip(np.hstack((0,gaps_ind)), np.hstack((gaps_ind, len(series))))

        # iterate over contiguous segments
        for starti, endi in gap_pairs:
            if (endi-starti) < 10:
                raise Exception("Data contains extremely small contiguous section {}[{}:{}]" \
                                    .format(channel.id,starti,endi))

            print "Sending contiguous data: {} to {}".format(starti, endi)
            # get contiguous segment
            data = series[starti:endi+1]
            # send contiguous region of data
            self._send_contiguous(channel=channel, data=data, period=period)



'''
description:
  The profile management system enables the user to easily create multiple profiles for easily
  switching between accounts or organizations. The user can also set/unset global settings which
  apply to all profiles.

  Settings are loaded in the following order, with each tier overriding the last:
    1. Global settings - global settings are loaded first
    2. Default profile - if set, the default profile is loaded
    3. Environment variables - e.g.: BLACKFYNN_API_TOKEN
    4. Command line arguments - e.g.: --profile=<name>

usage:
  bf profile [options] [<command>] [<args>...]

commands: 
  create <name>              Create new profile
  delete <name>              Delete profile
  set-default (none|<name>)  Set default profile (or none).
  list [-c --contents]       List profiles (and optionally their contents), also highlights
                             default profile.
  show (global|<name>)       Show ALL settings for a profile

specific profile options:
  -c --contents              List profiles with contents. Also lists global settings, if any
  -f --force                 Attempt action without promptin for confirmation. Use with care

global options:
  -h --help                  Show help
  --profile=<name>          Use specified profile (instead of default)

advanced commands:
  set (global|<name>) <key> <value>   Set key/value pair for given profile, or globally 
  unset (global|<name>) <key>         Unset key/value pair for given profile, or globally
  keys                                List all available keys and their default values
'''

import os
from docopt import docopt

from blackfynn import settings

def main():
    args = docopt(__doc__)

    if not os.path.exists(settings.config_file):
        setup_assistant()
        
    elif args['<command>'] == 'create'      : create_profile(args['<args>']) 
    elif args['<command>'] == 'delete'      : delete_profile(args['<args>'], args['--force'])
    elif args['<command>'] == 'set-default' : set_default(args['<args>'])
    elif args['<command>'] == 'list' or args['--contents']: list_profiles(args['--contents'])

    elif args['<command>'] == 'show' : show_profile(args['<args>'])
    
    elif args['<command>'] == 'set'         : set_key(args['<args>'],args['--force'])
    elif args['<command>'] == 'unset'       : unset_key(args['<args>'],args['--force'])
    elif args['<command>'] == 'keys'        : list_keys()
    
    elif args['<command>'] in [None, 'help']:
        print(__doc__.strip('\n'))

    else:
        invalid_usage()

    with open(settings.config_file,'w') as configfile:
        settings.config.write(configfile)

def setup_assistant():
    settings.config.clear()
    
    print("Blackfynn profile setup assistant")

    settings.config['global'] = {'default_profile' : 'none'}

    print("Create a profile:")
    create_profile()

    print("Setup complete. Enter 'bf profile help' for available commands and actions")
        
#User commands
#=======================================================
def create_profile(name=[]):
    if len(name) == 0:
        name = raw_input('  Profile name [default]: ') or 'default'

    elif len(name) == 1:
        name = name[0]

    else:
        invalid_usage()
    
    if name in settings.config:
        if name == 'global':
            print("Profile name 'global' reserved for system. Please try a different name")
        else:
            print("Profile '{}' already exists".format(name))
    else:
        print("Creating profile '{}'".format(name))
        settings.config[name] = {}

        settings.config[name]['api_token']  = raw_input('  API token: ')
        settings.config[name]['api_secret'] = raw_input('  API secret: ')
            
        if settings.config['global']['default_profile'] == 'none':
            settings.config['global']['default_profile'] = name
            print("Default profile: {}".format(name))
            
        else:
            set_default = raw_input("Would you like to set '{}' as default (Y/n)? ".format(name,settings.config['global']['default_profile'])).lower() in ['y','yes']
            if set_default:
                settings.config['global']['default_profile'] = name
                print("Default profile: {}".format(name))

def delete_profile(name,f):
    if len(name) != 1: invalid_usage()
    else: name = name[0]

    if valid_name(name):
        if not f:
            f = raw_input("Delete profile '{}' (Y/n)? ".format(name)).lower() in ['y','yes']
            
        if f:
            print("Deleting profile '{}'".format(name))
            settings.config.remove_section(name)

            if settings.config['global']['default_profile'] == name:
                settings.config['global']['default_profile'] = 'none'
                print("\033[31m* Warning: default profile unset. Use 'bf profile set-default <name>' to set a new default\033[0m")
        else:
            print('abort')
                
def list_profiles(contents):
    '''
    Lists all profiles
    '''
    print('Profiles:')
    for section in settings.config.sections():
        if section not in ['global']:
            if section == settings.config['global']['default_profile']:
                print('* \033[32m{}\033[0m'.format(section))
            else:
                print('  {}'.format(section))
            if contents:
                print_profile(section,4)
    if contents:
        if len(settings.config['global'].items()) > 1:
            print('Global Settings:')
            print_profile('global',2)


def set_default(name):
    if len(name) != 1: invalid_usage()
    else: name = name[0]
    
    if name == 'none':
        print("Default profile unset. Using global settings and environment variables")
        settings.config['global']['default_profile'] = 'none'
    elif valid_name(name):
        print("Default profile: {}".format(name))
        settings.config['global']['default_profile'] = name

def show_profile(name):
    if len(name) != 1: invalid_usage()
    else: name = name[0]

    if name == 'global' or valid_name(name):
        print('{} contents:'.format(name))
        print_profile(name,2,True)
         
#Advanced commands
#=======================================================
def set_key(args,force):
    if len(args) != 3: invalid_usage()

    name,key,value = args

    if not key in settings.defaults:
        print("Invalid key: '{}'\n see 'bf profile keys' for available keys".format(key))
        return

    if not name in settings.config:
        print("Profile '{}' does not exist".format(name))
        return

    if not force and key in settings.config[name]:
        force = raw_input("{}: {} already set. Overwrite (Y/n)? ".format(name,key)).lower() in ['y','yes']
    else: force = True
        
    if force:
        print("{}: {}={}".format(name,key,value))
        settings.config[name][key] = value

def unset_key(args,force):
    if len(args) != 2: invalid_usage()

    name,key = args

    if not name in settings.config:
        print("Profile '{}' does not exist".format(name))
        return

    if not key in settings.config[name]:
        print("{}: {} not set".format(name,key))
        return
        
    if not force:
        if key in settings.config[name]: force = raw_input("{}: Unset {} (Y/n)? ".format(name,key)).lower() in ['y','yes']

    if force:
        print("{}: {} unset".format(name,key))
        settings.config[name].pop(key)

def list_keys():
    print('Keys and default values:')
    for key, value in sorted(settings.defaults.items()):
        print('  {} : {}'.format(key, value))


#Helper functions
#=======================================================
def invalid_usage():
    print('Invalid usage. See bf profile help for available commands')
    exit()

def valid_name(name):
    if name not in settings.config or name=='global':
        print("Profile '{}' does not exist".format(name))
        return False
    return True

def print_profile(name,indent=0,show_all=False):
    if show_all:
        key_len = 0
        for key in settings.profiles[name].keys(): key_len = max(key_len, len(key))

        for key, value in sorted(settings.profiles[name].items()):
            if key != 'default_profile':
                if key in settings.config[name] and name != 'global':
                    print(' '*indent + '{:{key_len}} : \033[32m{}\033[0m ({})'.format(key,value,name,key_len=key_len))
                elif key in settings.config['global']:
                    print(' '*indent + '{:{key_len}} : \033[34m{}\033[0m (global)'.format(key,value,key_len=key_len))
                else:
                    print(' '*indent + '{:{key_len}} : \033[0m{}\033[0m'.format(key,value,key_len=key_len))

    else:
        key_len = 0
        for key in settings.config[name].keys(): key_len = max(key_len,len(key))

        for key, value in sorted(settings.config[name].items()):
            if key != 'default_profile':
                print(' '*indent + '{:{key_len}} : {}'.format(key,value,key_len=key_len))



'''
usage:
  bf props [options] <item> get [<key>]
  bf props [options] <item> set <key> <value> [--type=<type>]
  bf props [options] <item> rm  <key>

property options:
  -c --category=<cat>      Category of property [default: blackfynn]
  -t --type=<type>         Data type; string, integer, double, boolean, date, user [default: string]

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt
from blackfynn.models import Property
from cli_utils import get_client, get_item

def get_prop(item, key, cat):
    prop = item.get_property(key, category=cat)
    if prop is None:
        print "Error: property key \'{k}\' not found (category={c})".format(k=key,c=cat)
        return
    return prop

def main():
    args = docopt(__doc__)

    bf = get_client()

    item = get_item(args['<item>'], bf)
    key  = args['<key>']
    cat  = args['--category']

    if args['get']:
        if key:
            prop = get_prop(item, key, cat)
            if prop is None: return
            print "{cat} / {key}: {value}".format(cat=prop.category, key=prop.key, value=prop.value)
            return
        for prop in item.properties:
            print " * {cat} / {key}: {value}".format(cat=prop.category, key=prop.key, value=prop.value)
    elif args['set']:
        val  = args['<value>']
        dtype = args['--type']
        if dtype not in Property._data_types:
            print "Error: Data type must be one of {}".format(Property._data_types)
            return
        item.insert_property(key, val, category=cat, data_type=dtype)
        print "SET {cat} / {key} = {value} ({type})".format(cat=cat, key=key, value=val, type=dtype)
    elif args['rm']:
        prop = get_prop(item, key, cat)
        if prop is None: return
        item.remove_property(prop.key, category=prop.category)
        print "REMOVED {cat} / {key}".format(cat=cat, key=key)



'''
usage:
  bf collaborators [options]
  bf share [options] <collaborators>...
  bf unshare [options] <collaborators>...

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt
import sys

from cli_utils import get_client, get_working_dataset, print_collaborator_edit_resp

def main():
    args = docopt(__doc__)

    bf = get_client()
    ds = get_working_dataset(bf)

    if args['collaborators']:
        resp = ds.collaborators

        if resp['users'] or resp['organizations']:
            print " - Users"
            for u in resp['users']:
                print "   - email:{} id:{}".format(u.email, u.id)

            print " - Organizations"
            for o in resp['organizations']:
                print "   - name:{} id:{}".format(o.name, o.id)
        else:
            print "  No collaborators."

    elif args['share']:
        ids = args['<collaborators>']

        resp = ds.add_collaborators(*ids)
        print_collaborator_edit_resp(resp)

    elif args['unshare']:
        ids = args['<collaborators>']

        resp = ds.remove_collaborators(*ids)
        print_collaborator_edit_resp(resp)



import os
import re
from pip.download import PipSession
from pip.req import parse_requirements
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

with open('blackfynn/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')

install_reqs = parse_requirements('requirements.txt', session=PipSession())
reqs = [str(ir.req) for ir in install_reqs]


here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name = "blackfynn",
    version = version,
    author = "Blackfynn, Inc.",
    author_email = "mark@blackfynn.com",
    description = "Python client for the Blackfynn Platform",
    long_description = long_description,
    packages=find_packages(),
    package_dir={'blackfynn': 'blackfynn'},
    setup_requires=['cython'],
    install_requires = reqs,
    python_requires='<3',
    entry_points = {
        'console_scripts': [
            'bf=blackfynn.cli.bf:blackfynn_cli',
        ]
    },
    license = "",
    keywords = "blackfynn client rest api",
    url = "http://app.blackfynn.io",
    download_url = "",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
    ],
)



# -*- coding: utf-8 -*-

import requests

# blackfynn-specific
from blackfynn import settings
from blackfynn.models import Dataset
from blackfynn.api.transfers import IOAPI
from blackfynn.api.compute import ComputeAPI
from blackfynn.api.ledger import LedgerAPI
from blackfynn.api.user import UserAPI
from blackfynn.api.timeseries import TimeSeriesAPI
from blackfynn.base import ClientSession
from blackfynn.api.core import (
    CoreAPI, SecurityAPI, OrganizationsAPI, SearchAPI
)
from blackfynn.api.data import (
    DatasetsAPI, PackagesAPI, FilesAPI, DataAPI, TabularAPI
)


class Blackfynn(object):
    """
    The client provides an interface to the Blackfynn platform, giving you the
    ability to retrieve, add, and manipulate data.

    Args:
        api_token (str, optional): User's API token
        host (str, optional): API address
        streaming_host (str, optional): Streaming API address

    Examples:
        Load the client library and initialize::

            from blackfynn import Blackfynn

            # note: no API token needed if environment variables are set
            bf = Blackfynn()

        Retrieve/modify items on the platform::

            # get some dataset
            ds = bf.get_dataset('my dataset')
            print "Dataset {} has ID={}".format(ds.name, ds.id)

            # list all first-level items in dataset
            for item in ds:
                print "Item: {}".format(item)

            # grab some data package
            pkg = bf.get("N:package:1234-1234-1234-1234")

            # modify a package's name
            pkg.name = "New Package Name"

            # add some property
            pkg.set_property('Room', 'ICU-123')

            # sync changes
            pkg.update()


    Note:
        To initialize your ``Blackfynn`` client without passing any arguments,
        ensure that your ``BLACKFYNN_API_TOKEN`` and ``BLACKFYNN_API_SECRET`` environment variables
        are properly set.

    """
    def __init__(self, profile=None, api_token=None, api_secret=None, host=None, streaming_host=None):

        #Set profile, if present
        settings.use_profile(profile)

        #Set variables from named arguments, or settings
        api_token      = api_token      if api_token      is not None else settings.api_token
        api_secret     = api_secret     if api_secret     is not None else settings.api_secret
        host           = host           if host           is not None else settings.api_host
        streaming_host = streaming_host if streaming_host is not None else settings.streaming_api_host

        if api_token  is None: raise Exception('Error: No API token found. Cannot connect to Blackfynn.')
        if api_secret is None: raise Exception('Error: No API secret found. Cannot connect to Blackfynn.')
        
        self.host = host
        self.streaming_host = streaming_host

        # direct interface to REST API.
        self._api = ClientSession(api_token=api_token, api_secret=api_secret, host=host, streaming_host=streaming_host)

        # account
        try:
            self._api.authenticate()
        except Exception as e:
            raise Exception("Unable to connect to Blackfynn using the provided API token.")

        self._api.register(
            CoreAPI,
            OrganizationsAPI,
            DatasetsAPI,
            FilesAPI,
            DataAPI,
            PackagesAPI,
            TimeSeriesAPI,
            TabularAPI,
            SecurityAPI,
            ComputeAPI,
            SearchAPI,
            IOAPI,
            LedgerAPI,
            UserAPI
        )

        self._api._context = self._api.organizations.get(self._api._organization)


    @property
    def context(self):
        """
        The current organizational context of the active client.
        """
        return self._api._context

    @property
    def profile(self):
        """
        The profile of the current active user.
        """
        return self._api.profile

    def organizations(self):
        """
        Return all organizations for user.
        """
        return self._api.organizations.get_all()

    def datasets(self):
        """
        Return all datasets for user for an organization (current context).
        """
        self._check_context()
        return self.context.datasets

    def get(self, id, update=True):
        """
        Get any DataPackage or Collection object by ID. 

        Args:
            id (str): The ID of the Blackfynn object.

        Returns:
            Object of type/subtype ``DataPackage`` or ``Collection``.
        """
        try:
            return self._api.core.get(id, update=update)
        except:
            print("Unable to retrieve object"\
                  "\n\nAcceptable objects for get() are:"\
                  "\n - DataPackages"\
                  "\n - Collections"\
                  "\n\nUse get_dataset() if trying to retrieve a dataset")

    def create_dataset(self, name):
        """
        Create a dataset under the active organization.

        Args:
            name (str): The name of the to-be-created dataset

        Returns:
            The created ``Dataset`` object

        """
        self._check_context()
        return self._api.datasets.create(Dataset(name))

    def get_dataset(self, name_or_id):
        """
        Get Dataset by name or ID. 

        Args:
            name_or_id (str): the name or the ID of the dataset

        Note:
            When using name, this method gnores case, spaces, hyphens, 
            and underscores such that these are equivelent: 

              - "My Dataset"
              - "My-dataset"
              - "mydataset"
              - "my_DataSet"
              - "mYdata SET"

        """
        try:
            return self._api.datasets.get(name_or_id)
        except:
            pass

        result = self._api.datasets.get_by_name_or_id(name_or_id)
        if result is None:
            raise Exception("No dataset matching name or ID '{}'.".format(name_or_id))
        return result

    def update(self, thing):
        """
        Update an item on the platform.

        Args:
            thing (object or str): the ID or object to update

        Example::

            my_eeg = bf.get('N:package:1234-1234-1234-1234')
            my_eeg.name = "New EEG Name"
            bf.update(my_eeg)

        Note:
            You can also call update directly on objects, for example::

                my_eeg.name = "New EEG Name"
                my_eeg.update()

        """ 
        return self._api.core.update(thing)

    def delete(self, *things):
        """
        Deletes items from the platform.

        Args:
            things (list): ID or object of items to delete

        Example::

            # delete a package by passing in object
            bf.delete(my_package)

            # delete a package by ID
            bf.delete('N:package:1234-1234-1234-1235')

            # delete many things (mix of IDs and objects)
            bf.delete(my_package, 'N:collection:1234-1234-1234-1235', another_collection)

        Note:
            When deleting ``Collection``, all child items will also be deleted.

        """
        return self._api.core.delete(*things)

    def move(self, destination, *things):
        """
        Moves item(s) from their current location to the destination.

        Args:
            destination (object or str): The ID of the destination. This must be of type
                type ``Collection`` or ``None``. If destination is ``None``, ``things``
                will be moved to the top of their containing ``Dataset``
            things (list): the IDs or objects to move.

        """
        r = self._api.data.move(destination, *things)

    def search(self, query, max_results=10):
        """
        Find an object on the platform. 

        Args:
            query (str): query string to perform search.
            max_results (int, optional): the number of results to return

        Example::

            # find some items belonging to patient123
            for result in  bf.search('patient123'):
                print "found:", result

        """
        return self._api.search.query(query, max_results=max_results)
        
    def _check_context(self):
        if self.context is None:
            raise Exception('Must set context before executing method.') 

    def __repr__(self):
        return "<Blackfynn user='{}' organization='{}'>".format(self.profile.email, self.context.name)



'''Describe a package or collection

usage:
  bf get [options] [--tree] [<item>]

options:
  --tree                    Recursively print the item's contents

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt

from cli_utils import get_client, get_working_dataset, get_item

def display(item, print_tree):
    if print_tree:
        item.print_tree()

    else:
        print ("  {}".format(item))

        if hasattr(item, 'items') and item.items:
            print "  [contents]"
            for i in item.items:
                print "     {}".format(i)

        if hasattr(item, 'channels') and item.channels:
            print "  [channels]"
            for ch in item.channels:
                print "     {} (id: {})".format(ch.name, ch.id)

def main():
    args = docopt(__doc__)

    bf = get_client()
    item = get_item(args['<item>'], bf) if args['<item>'] else get_working_dataset(bf)

    display(item, args['--tree'])
